# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The old pylife-1.x rainflow counting API

In order to not to break existing code, the old pylife-1.x API is still in
place as wrappers around the new API. Using it is strongly discouraged. It will
be deprecated and eventually removed.
"""

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd

from .threepoint import ThreePointDetector
from .fkm import FKMDetector
from .recorders import FullRecorder


class AbstractRainflowCounter:
    def __init__(self):
        self._recorder = FullRecorder()

    @property
    def loops_from(self):
        return self._recorder.values_from

    @property
    def loops_to(self):
        return self._recorder.values_to

    def residuals(self):
        return self._detector.residuals

    def get_rainflow_matrix(self, bins):
        return self._recorder.histogram_numpy(bins)

    def get_rainflow_matrix_frame(self, bins):
        return pd.DataFrame(self._recorder.histogram(bins))


class RainflowCounterThreePoint(AbstractRainflowCounter):
    def __init__(self):
        super().__init__()
        self._detector = ThreePointDetector(recorder=self._recorder)

    def process(self, samples):
        self._detector.process(samples)
        return self


class RainflowCounterFKM(AbstractRainflowCounter):
    def __init__(self):
        super().__init__()
        self._detector = FKMDetector(recorder=self._recorder)

    def process(self, samples):
        self._detector.process(samples)
        return self



# Copyright (c) 2019-2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .elementary import Elementary

import pandas as pd
import numpy as np
from scipy import optimize
import warnings


class MaxLikeInf(Elementary):
    """Maximum likelihood procedure estimating the ``SD_50`` and ``TS`` and ``ND_50``.

    Only the values describing the infinite lifetime (load endurance limit
    ``SD_50`` and load endurance scatter ``TS``) are calculated by maximum
    likelihood.  The slope ``k_1`` and ``TN`` are taken from the
    :class:`Elementary` calculation.

    The load cycle endurance ``ND_50`` is computed by intersecting load
    endurance limit line with the line of slope ``k_1``.

    """
    def _specific_analysis(self, wc):
        SD, TS = self.__max_likelihood_inf_limit()

        wc['SD'] = SD
        wc['TS'] = TS
        if not np.isnan(wc['ND']):
            wc['ND'] = self._transition_cycles(SD)

        return wc

    def __max_likelihood_inf_limit(self):

        infinite_fractures = self._fd.infinite_zone.loc[self._fd.infinite_zone.fracture]
        infinite_runouts = self._fd.infinite_zone.loc[~self._fd.infinite_zone.fracture]
        infinite_fractured_loads = np.unique(infinite_fractures.load.values)
        infinite_runout_loads = np.unique(infinite_runouts.load.values)
        infinite_mixed_loads = np.intersect1d(infinite_runout_loads, infinite_fractured_loads)

        def fail_if_less_than_two_mixed_levels():
            if len(infinite_mixed_loads) < 2:
                raise ValueError("MaxLikeHood: need at least two mixed load levels.")

        def fail_if_less_than_three_fractures():
            if infinite_fractures.shape[0] < 3 or len(infinite_fractured_loads) < 2:
                raise ValueError("MaxLikeHood: need at least three fractures on two load levels.")


        fail_if_less_than_two_mixed_levels()
        fail_if_less_than_three_fractures()

        SD_start = self._fd.finite_infinite_transition
        TS_start = 1.2

        var_opt = optimize.fmin(lambda p: -self._lh.likelihood_infinite(p[0], p[1]),
                                [SD_start, TS_start], disp=False, full_output=True)

        SD_50 = var_opt[0][0]
        TS = var_opt[0][1]

        return SD_50, TS


class MaxLikeFull(Elementary):
    """Maximum likelihood procedure estimating all parameters.

    Maximum likelihood is a method of estimating the parameters of a
    distribution model by maximizing a likelihood function, so that under the
    assumed statistical model the observed data is most probable.  This
    procedure consists of estimating the curve parameters, where some of these
    parameters may be fixed by the user. The remaining parameters are then
    fitted to produce the best possible outcome.

    https://en.wikipedia.org/wiki/Maximum_likelihood_estimation

    Parameters
    ----------
    fixed_parameters : dict
        Dictionary of fixed parameters

    """

    def _specific_analysis(self, wc, fixed_parameters={}):
        return pd.Series(self.__max_likelihood_full(wc, fixed_parameters))

    def __max_likelihood_full(self, initial_wcurve, fixed_prms):

        def warn_and_fix_if_no_runouts():
            nonlocal fixed_prms
            if self._fd.num_runouts == 0:
                warnings.warn(UserWarning("MaxLikeHood: no runouts are present in fatigue data. "
                                          "Proceeding with SD = 0 and TS = 1 as fixed parameters. "
                                          "This is NOT a standard evaluation!"))
                fixed_prms = fixed_prms.copy()
                fixed_prms.update({'SD': 0.0, 'TS': 1.0})

        def fail_if_less_than_three_fractures():
            if self._fd.num_fractures < 3 or len(self._fd.fractured_loads) < 2:
                raise ValueError("MaxLikeHood: need at least three fractures on two load levels.")

        def warn_and_fix_if_less_than_two_mixed_levels():
            nonlocal fixed_prms
            if len(self._fd.mixed_loads) < 2 and self._fd.num_runouts > 0 and self._fd.num_fractures > 0:
                warnings.warn(UserWarning("MaxLikeHood: less than two mixed load levels in fatigue data."
                                          "Proceeding by setting a predetermined scatter from the standard Wöhler curve."))
                fixed_prms = fixed_prms.copy()
                TN, TS = self._pearl_chain_method()
                fixed_prms.update({'TS': TS})

        fail_if_less_than_three_fractures()
        warn_and_fix_if_no_runouts()
        warn_and_fix_if_less_than_two_mixed_levels()

        p_opt = initial_wcurve.to_dict()
        for k in fixed_prms:
            p_opt.pop(k)

        if not p_opt:
            raise AttributeError('You need to leave at least one parameter empty!')
        var_opt = optimize.fmin(
            self.__likelihood_wrapper, [*p_opt.values()],
            args=([*p_opt], fixed_prms),
            full_output=True,
            disp=False,
            maxiter=1e4,
            maxfun=1e4,
        )
        res = {}
        res.update(fixed_prms)
        res.update(zip([*p_opt], var_opt[0]))

        return self.__make_parameters(res)

    def __make_parameters(self, params):
        params['SD'] = np.abs(params['SD'])
        params['TS'] = np.abs(params['TS'])
        params['k_1'] = np.abs(params['k_1'])
        params['ND'] = np.abs(params['ND'])
        params['TN'] = np.abs(params['TN'])
        return params

    def __likelihood_wrapper(self, var_args, var_keys, fix_args):
        ''' 1) Finds the start values to be optimized. The rest of the paramters are fixed by the user.
            2) Calls function mali_sum_lolli to calculate the maximum likelihood of the current
            variable states.
        '''
        args = {}
        args.update(fix_args)
        args.update(zip(var_keys, var_args))
        args = self.__make_parameters(args)

        return -self._lh.likelihood_total(args['SD'], args['TS'], args['k_1'], args['ND'], args['TN'])



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import numpy as np
import pandas as pd

from pylife.core.broadcaster import Broadcaster


foo_bar_series = pd.Series({'foo': 1.0, 'bar': 2.0})
foo_bar_series_twice_in_frame = pd.DataFrame([foo_bar_series, foo_bar_series])

series_named_index = foo_bar_series.copy()
series_named_index.index.name = 'idx1'

foo_bar_frame = pd.DataFrame({'foo': [1.0, 1.5], 'bar': [2.0, 1.5]})


def test_broadcast_series_to_array():
    param, obj = Broadcaster(foo_bar_series).broadcast([1.0, 2.0])

    pd.testing.assert_series_equal(param, pd.Series([1.0, 2.0]))
    pd.testing.assert_frame_equal(foo_bar_series_twice_in_frame, obj)


def test_broadcast_frame_to_array_match():
    param, obj = Broadcaster(foo_bar_frame).broadcast([1.0, 2.0])

    np.testing.assert_array_equal(param, [1.0, 2.0])
    pd.testing.assert_frame_equal(foo_bar_frame, obj)


def test_broadcast_frame_to_array_mismatch():
    with pytest.raises(ValueError, match=r"Dimension mismatch. "
                       "Cannot map 3 value array-like to a 2 element DataFrame signal."):
        Broadcaster(foo_bar_frame).broadcast([1.0, 2.0, 3.0])


def test_broadcast_series_to_scalar():
    param, obj = Broadcaster(foo_bar_series).broadcast(1.0)

    assert param == 1.0
    pd.testing.assert_series_equal(foo_bar_series, obj)


def test_broadcast_frame_to_scalar():
    param, obj = Broadcaster(foo_bar_frame).broadcast(1.0)

    expected_param = pd.Series([1.0, 1.0], index=foo_bar_frame.index)
    pd.testing.assert_series_equal(expected_param, param)
    pd.testing.assert_frame_equal(foo_bar_frame, obj)


def test_broadcast_series_index_named_to_series_index_named():
    series = pd.Series([5.0, 6.0], index=pd.Index(['x', 'y'], name='idx2'))
    param, obj = Broadcaster(series_named_index).broadcast(series)

    expected_param = pd.Series({
        ('foo', 'x'): 5.0,
        ('foo', 'y'): 6.0,
        ('bar', 'x'): 5.0,
        ('bar', 'y'): 6.0
    })
    expected_obj = pd.Series({
        ('foo', 'x'): 1.0,
        ('foo', 'y'): 1.0,
        ('bar', 'x'): 2.0,
        ('bar', 'y'): 2.0
    })
    expected_obj.index.names = ['idx1', 'idx2']
    expected_param.index.names = ['idx1', 'idx2']

    pd.testing.assert_series_equal(expected_param, param)
    pd.testing.assert_series_equal(expected_obj, obj)


def test_broadcast_series_index_named_to_series_index_none():
    series = pd.Series([5.0, 6.0], index=pd.Index([3, 4]))
    param, obj = Broadcaster(series_named_index).broadcast(series)

    expected_param = pd.Series({
        ('foo', 3): 5.0,
        ('foo', 4): 6.0,
        ('bar', 3): 5.0,
        ('bar', 4): 6.0
    })
    expected_obj = pd.Series({
        ('foo', 3): 1.0,
        ('foo', 4): 1.0,
        ('bar', 3): 2.0,
        ('bar', 4): 2.0
    })
    expected_obj.index.names = ['idx1', None]
    expected_param.index.names = ['idx1', None]

    pd.testing.assert_series_equal(expected_param, param)
    pd.testing.assert_series_equal(expected_obj, obj)


def test_broadcast_series_index_none_to_series_index_none():
    series = pd.Series([1.0, 2.0], index=pd.Index([3, 4]))
    param, obj = Broadcaster(foo_bar_series).broadcast(series)

    expected = pd.DataFrame([foo_bar_series, foo_bar_series], index=series.index)
    pd.testing.assert_series_equal(series, param)
    pd.testing.assert_frame_equal(expected, obj)


def test_broadcast_series_index_none_to_series_index_none_no_string_index():
    series = pd.Series([1.0, 2.0], index=pd.Index([3, 4]))
    obj = foo_bar_series.copy()
    obj.index = pd.Index([1, 2])
    param, obj = Broadcaster(obj).broadcast(series)

    expected = pd.DataFrame([foo_bar_series, foo_bar_series],
                            index=series.index)
    expected.columns = [1, 2]
    pd.testing.assert_series_equal(series, param)
    pd.testing.assert_frame_equal(expected, obj)


def test_broadcast_series_index_none_to_series_index_named():
    series = pd.Series([1.0, 2.0], index=pd.Index([3, 4], name='idx2'))
    foo_bar = foo_bar_series.copy()
    foo_bar.index.name = None
    param, obj = Broadcaster(foo_bar).broadcast(series)

    expected = pd.DataFrame([foo_bar_series, foo_bar_series], index=series.index)
    pd.testing.assert_series_equal(series, param)
    pd.testing.assert_frame_equal(expected, obj)


def test_broadcast_series_to_frame_2_elements_index_none():
    df = pd.DataFrame({
        'a': [1, 3],
        'b': [2, 4]
    }, index=['x', 'y'])

    param, obj = Broadcaster(foo_bar_series).broadcast(df)

    expected_obj = pd.DataFrame({
        'foo': [1.0, 1.0], 'bar': [2.0, 2.0]
    }, index=['x', 'y'])

    pd.testing.assert_frame_equal(param, df)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_series_to_frame_3_elements_index_none():
    df = pd.DataFrame({
        'a': [1, 3, 5],
        'b': [2, 4, 6]
    }, index=['x', 'y', 'z'])

    param, obj = Broadcaster(foo_bar_series).broadcast(df)

    expected_obj = pd.DataFrame({
        'foo': [1.0, 1.0, 1.0], 'bar': [2.0, 2.0, 2.0],
    }, index=['x', 'y', 'z'])

    pd.testing.assert_frame_equal(param, df)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_series_to_series_same_single_index():
    series = pd.Series([1, 3], index=pd.Index(['x', 'y'], name='iname1'), name='src')

    foo_bar = pd.Series([1, 2], index=pd.Index(['x', 'y'], name='iname1'), name='dst')

    param, obj = Broadcaster(foo_bar).broadcast(series)
    pd.testing.assert_series_equal(param, series)
    pd.testing.assert_series_equal(obj, foo_bar)


def test_broadcast_series_to_series_different_single_index_name():
    series = pd.Series([1, 3], index=pd.Index(['x', 'y'], name='iname1'), name='dest')

    foo_bar = pd.Series([1, 2], index=pd.Index([1, 2], name='srcname'), name='src')

    expected_index = pd.MultiIndex.from_tuples([(1, 'x'), (1, 'y'), (2, 'x'), (2, 'y')], names=['srcname', 'iname1'])
    expected_obj = pd.Series([1, 1, 2, 2], name='src', index=expected_index)
    expected_param = pd.Series([1, 3, 1, 3], name='dest', index=expected_index)

    param, obj = Broadcaster(foo_bar).broadcast(series)

    pd.testing.assert_series_equal(param, expected_param)
    pd.testing.assert_series_equal(obj, expected_obj)


def test_broadcast_frame_to_series_one_common():
    series = pd.Series(3.0, index=pd.Index([1,2], name='bar'))
    df = pd.DataFrame(
        {'col': np.arange(6)},
        index=pd.MultiIndex.from_product([pd.Index([1, 2, 3]), series.index])
    )

    param, obj = Broadcaster(df).broadcast(series)
    Broadcaster(df.col).broadcast(series)

    pd.testing.assert_series_equal(param, pd.Series(3.0, index=df.index))


def test_broadcast_frame_to_frame_same_single_index():
    df = pd.DataFrame({
        'a': [1, 3],
        'b': [2, 4]
    }, index=pd.Index(['x', 'y'], name='iname1'))

    foo_bar = pd.DataFrame({'foo': [1, 2], 'bar': [3, 4]}, index=pd.Index(['x', 'y'], name='iname1'))

    param, obj = Broadcaster(foo_bar).broadcast(df)
    pd.testing.assert_frame_equal(param, df)
    pd.testing.assert_frame_equal(obj, foo_bar)


def test_broadcast_frame_to_frame_same_multi_index():
    index = pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('y', 1), ('y', 2)], names=['iname1', 'iname2'])
    df = pd.DataFrame({
        'a': [1, 3, 5, 7],
        'b': [2, 4, 6, 8]
    }, index=index)

    foo_bar = pd.DataFrame({'foo': [1, 2, 3, 4], 'bar': [3, 4, 5, 6]}, index=index)

    param, obj = Broadcaster(foo_bar).broadcast(df)
    pd.testing.assert_frame_equal(param, df)
    pd.testing.assert_frame_equal(obj, foo_bar)


def test_broadcast_frame_to_frame_same_single_index_name_different_elements():
    df = pd.DataFrame({
        'a': [1, 3, 5],
        'b': [2, 4, 6]
    }, index=pd.Index(['x', 'y', 'z'], name='iname1'))

    foo_bar = pd.DataFrame({
        'foo': [1, 2, 3],
        'bar': [3, 4, 5]
    }, index=pd.Index(['y', 'a', 'x'], name='iname1'))

    param, obj = Broadcaster(foo_bar).broadcast(df)

    expected_param = pd.DataFrame({
        'a': [1, 3, 5, np.nan],
        'b': [2, 4, 6, np.nan]
    }, index=pd.Index(['x', 'y', 'z', 'a'], name='iname1')).sort_index()

    expected_obj = pd.DataFrame({
        'foo': [1, 2, 3, np.nan],
        'bar': [3, 4, 5, np.nan]
    }, index=pd.Index(['y', 'a', 'x', 'z'], name='iname1')).sort_index()

    pd.testing.assert_frame_equal(param.sort_index(), expected_param)
    pd.testing.assert_frame_equal(obj.sort_index(), expected_obj)


def test_broadcast_frame_to_frame_same_multi_index_name_different_elements():
    df = pd.DataFrame({
        'a': [1, 3, 5, 7, 9, 11, 13, 15, 17],
        'b': [2, 4, 6, 8, 10, 12, 14, 16, 18]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('x', 3),
                                        ('y', 1), ('y', 2), ('y', 3),
                                        ('z', 1), ('z', 2), ('z', 3)], names=['iname1', 'iname2']))

    foo_bar = pd.DataFrame({
        'foo': [1, 2, 3, 4, 5, 6, 7, 8],
        'bar': [3, 4, 5, 6, 7, 8, 9, 0]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('x', 3), ('x', 4),
                                        ('y', 1), ('y', 2), ('y', 3), ('y', 4)], names=['iname1', 'iname2']))

    param, obj = Broadcaster(foo_bar).broadcast(df)

    expected_param = pd.DataFrame({
        'a': [1, 3, 5, np.nan, 7, 9, 11, np.nan, 13, 15, 17],
        'b': [2, 4, 6, np.nan, 8, 10, 12, np.nan, 14, 16, 18]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('x', 3), ('x', 4),
                                        ('y', 1), ('y', 2), ('y', 3), ('y', 4),
                                        ('z', 1), ('z', 2), ('z', 3)], names=['iname1', 'iname2']))

    expected_obj = pd.DataFrame({
        'foo': [1, 2, 3, 4, 5, 6, 7, 8, np.nan, np.nan, np.nan],
        'bar': [3, 4, 5, 6, 7, 8, 9, 0, np.nan, np.nan, np.nan]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('x', 3), ('x', 4),
                                        ('y', 1), ('y', 2), ('y', 3), ('y', 4),
                                        ('z', 1), ('z', 2), ('z', 3)], names=['iname1', 'iname2']))

    pd.testing.assert_frame_equal(param.sort_index(), expected_param.sort_index())
    pd.testing.assert_frame_equal(obj.sort_index(), expected_obj.sort_index())


def test_broadcast_frame_to_frame_different_single_index_name():
    df = pd.DataFrame({
        'a': [1, 3],
        'b': [2, 4]
    }, index=pd.Index(['x', 'y'], name='iname1'))

    foo_bar = pd.DataFrame({'foo': [1, 2], 'bar': [3, 4]}, index=pd.Index([1, 2], name='srcname'))

    expected_obj = pd.DataFrame({
        'foo': [1, 1, 2, 2],
        'bar': [3, 3, 4, 4],
    }, index=pd.MultiIndex.from_tuples([(1, 'x'), (1, 'y'), (2, 'x'), (2, 'y')], names=['srcname', 'iname1']))

    expected_param = pd.DataFrame({
        'a': [1, 3, 1, 3],
        'b': [2, 4, 2, 4]
    }, index=pd.MultiIndex.from_tuples([(1, 'x'), (1, 'y'), (2, 'x'), (2, 'y')], names=['srcname', 'iname1']))

    param, obj = Broadcaster(foo_bar).broadcast(df)

    pd.testing.assert_frame_equal(param, expected_param)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_frame_to_frame_different_multi_index_name():
    df = pd.DataFrame({
        'a': [1, 3, 5, 7],
        'b': [2, 4, 6, 8]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('y', 1), ('y', 2)], names=['iname1', 'iname2']))

    foo_bar = pd.DataFrame({
        'foo': [1, 2, 3, 4],
        'bar': [3, 4, 5, 6]
    }, index=pd.MultiIndex.from_tuples([('a', 10), ('a', 20), ('b', 10), ('b', 20)], names=['srcname1', 'srcname2']))

    param, obj = Broadcaster(foo_bar).broadcast(df)

    expected_obj = pd.DataFrame({
        'foo': [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4],
        'bar': [3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 10, 'x', 1), ('a', 10, 'x', 2), ('a', 10, 'y', 1), ('a', 10, 'y', 2),
        ('a', 20, 'x', 1), ('a', 20, 'x', 2), ('a', 20, 'y', 1), ('a', 20, 'y', 2),
        ('b', 10, 'x', 1), ('b', 10, 'x', 2), ('b', 10, 'y', 1), ('b', 10, 'y', 2),
        ('b', 20, 'x', 1), ('b', 20, 'x', 2), ('b', 20, 'y', 1), ('b', 20, 'y', 2)
    ], names=['srcname1', 'srcname2', 'iname1', 'iname2']))

    expected_param = pd.DataFrame({
        'a': [1, 3, 5, 7, 1, 3, 5, 7, 1, 3, 5, 7, 1, 3, 5, 7],
        'b': [2, 4, 6, 8, 2, 4, 6, 8, 2, 4, 6, 8, 2, 4, 6, 8]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 10, 'x', 1), ('a', 10, 'x', 2), ('a', 10, 'y', 1), ('a', 10, 'y', 2),
        ('a', 20, 'x', 1), ('a', 20, 'x', 2), ('a', 20, 'y', 1), ('a', 20, 'y', 2),
        ('b', 10, 'x', 1), ('b', 10, 'x', 2), ('b', 10, 'y', 1), ('b', 10, 'y', 2),
        ('b', 20, 'x', 1), ('b', 20, 'x', 2), ('b', 20, 'y', 1), ('b', 20, 'y', 2)
    ], names=['srcname1', 'srcname2', 'iname1', 'iname2']))

    pd.testing.assert_frame_equal(param, expected_param)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_frame_to_frame_different_multi_index_name_drop_level():
    df = pd.DataFrame({
        'a': [1, 3, 5, 7],
        'b': [2, 4, 6, 8]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('y', 1), ('y', 2)], names=['iname1', 'iname2']))

    foo_bar = pd.DataFrame({
        'foo': [1, 2, 3, 4],
        'bar': [3, 4, 5, 6]
    }, index=pd.MultiIndex.from_tuples([('a', 10), ('a', 20), ('b', 10), ('b', 20)], names=['srcname1', 'srcname2']))

    param, obj = Broadcaster(foo_bar).broadcast(df, droplevel=['srcname2'])

    expected_obj = pd.DataFrame({
        'foo': [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4],
        'bar': [3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 10, 'x', 1), ('a', 10, 'x', 2), ('a', 10, 'y', 1), ('a', 10, 'y', 2),
        ('a', 20, 'x', 1), ('a', 20, 'x', 2), ('a', 20, 'y', 1), ('a', 20, 'y', 2),
        ('b', 10, 'x', 1), ('b', 10, 'x', 2), ('b', 10, 'y', 1), ('b', 10, 'y', 2),
        ('b', 20, 'x', 1), ('b', 20, 'x', 2), ('b', 20, 'y', 1), ('b', 20, 'y', 2)
    ], names=['srcname1', 'srcname2', 'iname1', 'iname2']))

    expected_param = pd.DataFrame({
        'a': [1, 3, 5, 7, 1, 3, 5, 7],
        'b': [2, 4, 6, 8, 2, 4, 6, 8]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 'x', 1), ('a', 'x', 2), ('a', 'y', 1), ('a', 'y', 2),
        ('b', 'x', 1), ('b', 'x', 2), ('b', 'y', 1), ('b', 'y', 2)
    ], names=['srcname1', 'iname1', 'iname2']))

    pd.testing.assert_frame_equal(param, expected_param)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_frame_to_frame_mixed_multi_index_name():
    df = pd.DataFrame({
        'a': [1, 3, 5, 7],
        'b': [2, 4, 6, 8]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('y', 1), ('y', 2)], names=['iname1', 'srcname2']))

    foo_bar = pd.DataFrame({
        'foo': [1, 2, 3, 4],
        'bar': [3, 4, 5, 6]
    }, index=pd.MultiIndex.from_tuples([('a', 1), ('b', 1), ('a', 2), ('b', 2)], names=['srcname1', 'srcname2']))

    param, obj = Broadcaster(foo_bar).broadcast(df)

    expected_obj = pd.DataFrame({
        'foo': [1, 1, 2, 2, 3, 3, 4, 4],
        'bar': [3, 3, 4, 4, 5, 5, 6, 6]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 1, 'x'), ('a', 1, 'y'),
        ('b', 1, 'x'), ('b', 1, 'y'),
        ('a', 2, 'x'), ('a', 2, 'y'),
        ('b', 2, 'x'), ('b', 2, 'y')
    ], names=['srcname1', 'srcname2', 'iname1']))

    expected_prm = pd.DataFrame({
        'a': [1, 5, 1, 5, 3, 7, 3, 7],
        'b': [2, 6, 2, 6, 4, 8, 4, 8]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 1, 'x'), ('a', 1, 'y'),
        ('b', 1, 'x'), ('b', 1, 'y'),
        ('a', 2, 'x'), ('a', 2, 'y'),
        ('b', 2, 'x'), ('b', 2, 'y')
    ], names=['srcname1', 'srcname2', 'iname1']))

    pd.testing.assert_frame_equal(param, expected_prm)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_frame_to_frame_mixed_multi_index_name_drop_level():
    df = pd.DataFrame({
        'a': [1, 3, 5, 7],
        'b': [2, 4, 6, 8]
    }, index=pd.MultiIndex.from_tuples([('x', 1), ('x', 2), ('y', 1), ('y', 2)], names=['iname1', 'srcname2']))

    foo_bar = pd.DataFrame({
        'foo': [1, 2, 3, 4],
        'bar': [3, 4, 5, 6]
    }, index=pd.MultiIndex.from_tuples([('a', 1), ('b', 1), ('a', 2), ('b', 2)], names=['srcname1', 'srcname2']))

    param, obj = Broadcaster(foo_bar).broadcast(df, droplevel=['srcname1'])

    expected_obj = pd.DataFrame({
        'foo': [1, 1, 2, 2, 3, 3, 4, 4],
        'bar': [3, 3, 4, 4, 5, 5, 6, 6]
    }, index=pd.MultiIndex.from_tuples([
        ('a', 1, 'x'), ('a', 1, 'y'),
        ('b', 1, 'x'), ('b', 1, 'y'),
        ('a', 2, 'x'), ('a', 2, 'y'),
        ('b', 2, 'x'), ('b', 2, 'y')
    ], names=['srcname1', 'srcname2', 'iname1']))

    expected_prm = pd.DataFrame({
        'a': [1, 5, 3, 7],
        'b': [2, 6, 4, 8]
    }, index=pd.MultiIndex.from_tuples([
        (1, 'x'), (1, 'y'),
        (2, 'x'), (2, 'y')
    ], names=['srcname2', 'iname1']))

    pd.testing.assert_frame_equal(param, expected_prm)
    pd.testing.assert_frame_equal(obj, expected_obj)


def test_broadcast_series_to_series_overlapping_interval_index():
    interval_index = pd.IntervalIndex.from_tuples([
        (0.0, 1.0), (1.0, 2.0), (1.5, 2.5)
    ], name='interval')

    obj = pd.Series([1, 2, 3], index=interval_index, name='series')

    operand = pd.Series([4, 5, 6], index=pd.RangeIndex(3, name='operand'), name='foo')

    prm, obj = Broadcaster(obj).broadcast(operand)

    expected_index = pd.MultiIndex.from_tuples([
        (pd.Interval(0.0, 1.0), 0),
        (pd.Interval(0.0, 1.0), 1),
        (pd.Interval(0.0, 1.0), 2),
        (pd.Interval(1.0, 2.0), 0),
        (pd.Interval(1.0, 2.0), 1),
        (pd.Interval(1.0, 2.0), 2),
        (pd.Interval(1.5, 2.5), 0),
        (pd.Interval(1.5, 2.5), 1),
        (pd.Interval(1.5, 2.5), 2),
    ], names=['interval', 'operand'])

    expected_obj = pd.Series([1, 1, 1, 2, 2, 2, 3, 3, 3], index=expected_index, name='series')
    expected_prm = pd.Series([4, 5, 6, 4, 5, 6, 4, 5, 6], index=expected_index, name='foo')

    pd.testing.assert_series_equal(obj, expected_obj)
    pd.testing.assert_series_equal(prm, expected_prm)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

from .abstract_load_collective import AbstractLoadCollective
from .load_histogram import LoadHistogram

import pandas as pd
import numpy as np

from pylife import PylifeSignal

@pd.api.extensions.register_dataframe_accessor('load_collective')
class LoadCollective(PylifeSignal, AbstractLoadCollective):
    """A Load collective.

    The usual use of this signal is to process hysteresis loop data from a
    rainflow recording.  Usually the keys ``from`` and ``to`` are used to
    describe the hysteresis loops.  Alternatively also the keys ``range`` and
    ``mean`` can be given.  In that case the frame is internally converted to
    ``from`` and ``to`` where the ``from`` values are the lower ones.
    """

    def _validate(self):
        if 'from' in self.keys() and 'to' in self.keys():
            return
        if 'range' in self.keys() and 'mean' in self.keys():
            fr = self._obj['mean'] - self._obj['range'] / 2.
            to = self._obj['mean'] + self._obj['range'] / 2.

            cycles = self._obj.get('cycles')

            self._obj = pd.DataFrame({
                'from': fr,
                'to': to
            }, index=self._obj.index)

            if cycles is not None:
                self._obj['cycles'] = cycles

            return
        raise AttributeError("Load collective needs either 'range'/'mean' or 'from'/'to' in column names.")

    @property
    def amplitude(self):
        """Calculate the amplitudes of the load collective.

        Returns
        -------
        amplitude : pd.Series
            The amplitudes of the load collective
        """
        fr = self._obj['from']
        to = self._obj['to']
        rng = np.abs(fr-to)

        return pd.Series(rng/2., name='amplitude', index=self._obj.index)

    @property
    def meanstress(self):
        """Calculate the mean load values of the load collective.

        Returns
        -------
        mean : pd.Series
            The mean load values of the load collective
        """
        fr = self._obj['from']
        to = self._obj['to']
        return pd.Series((fr+to)/2., name='meanstress')

    @property
    def R(self):
        """Calculate the R values of the load collective.

        Returns
        -------
        R : pd.Series
            The R values of the load collective
        """
        res = (self.lower / self.upper).fillna(0.0)
        res.name = 'R'
        return res

    @property
    def upper(self):
        """Calculate the upper load values of the load collective.

        Returns
        -------
        upper : pd.Series
            The upper load values of the load collective
        """
        res = self._obj.loc[:, ['from', 'to']].max(axis=1)
        res.name = 'upper'
        return res

    @property
    def lower(self):
        """Calculate the lower load values of the load collective.

        Returns
        -------
        lower : pd.Series
            The lower load values of the load collective
        """
        res = self._obj.loc[:, ['from', 'to']].min(axis=1)
        res.name = 'lower'
        return res

    @property
    def cycles(self):
        """The cycles of each member of the collective is 1.0. when no cycles are given

        This is for compatibility with :class:`~pylife.stress.pylife.stress.LoadHistogram`
        """
        if 'cycles' in self._obj.keys():
            return self._obj.cycles

        return pd.Series(1.0, name='cycles', index=self._obj.index)

    def scale(self, factors):
        """Scale the collective.

        Parameters
        ----------
        factors : scalar or :class:`pandas.Series`
            The factor(s) to scale the collective 'from' and 'to' with.

        Returns
        -------
        scaled : ``LoadHistogram``
            The scaled histogram.
        """
        factors, obj = self.broadcast(factors)
        obj[['from', 'to']] = obj[['from', 'to']].multiply(factors, axis=0)
        return obj.load_collective

    def shift(self, diffs):
        """Shift the collective.

        Parameters
        ----------
        diffs : scalar or :class:`pandas.Series`
            The diff(s) to shift the collective by.

        Returns
        -------
        shifted : ``LoadHistogram``
            The shifted histogram.
        """
        diffs, obj = self.broadcast(diffs)
        obj[['from', 'to']] = obj[['from', 'to']].add(diffs, axis=0)
        return obj.load_collective

    def range_histogram(self, bins, axis=None):
        """Calculate the histogram of range values along a given axis.

        Parameters
        ----------
        bins : int, sequence of scalars or pd.IntervalIndex
            The bins of the histogram to be calculated

        Returns
        -------
        range histogram : :class:`~pylife.pylife.stress.LoadHistogram`

        axis : str, optional
            The index axis along which the histogram is calculated. If missing
            the histogram is calculated over the whole collective.
        """
        def make_histogram(group):
            cycles, intervals = np.histogram(group * 2., bins)
            idx = pd.IntervalIndex.from_breaks(intervals, name='range')
            return pd.Series(cycles, index=idx, name='cycles')

        if isinstance(bins, pd.IntervalIndex):
            bins = np.append(bins.left[0], bins.right)

        if axis is None:
            return LoadHistogram(make_histogram(self.amplitude))

        result = pd.Series(self.amplitude
                           .groupby(self._obj.index.droplevel(axis).names)
                           .apply(make_histogram), name='cycles')

        return LoadHistogram(result)



# Copyright (c) 2020-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

import numpy as np
from h5py.h5t import string_dtype

from .exceptions import *
from .vmap_dataset import VMAPDataset


class VMAPUnit(VMAPDataset):
    def __init__(self, identifier, si_scale, si_shift, unit_symbol, unit_quantity):
        super().__init__(identifier)
        self._si_scale = si_scale
        self._si_shift = si_shift
        self._unit_symbol = unit_symbol
        self._unit_quantity = unit_quantity

    @property
    def attributes(self):
        if self._identifier is None:
            raise (APIUseError("Need to set_identifier() before requesting the attributes."))
        return self._identifier, self._si_scale, self._si_shift, self._unit_symbol, self._unit_quantity

    @property
    def dtype(self):
        dt_type = np.dtype({"names": ["myIdentifier", "mySIScale", "mySIShift", "myUnitSymbol", "myUnitQuantity"],
                            "formats": ['<i4', '<f8', '<f8', string_dtype(), string_dtype()]})
        return dt_type

    @property
    def dataset_name(self):
        return 'UNITSYSTEM'



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import pytest

import numpy as np
import pandas as pd
import numpy.testing as testing

import pylife.strength.meanstress as MST


def goodman_signal_sm():
    Sm = np.array([-4., -2.,   -1., 0., 0.4, 2./3., 7./6.])
    Sa = np.array([ 2.,  2., 3./2., 1., 0.8, 2./3., 7./12.])
    return pd.DataFrame({'mean': Sm, 'range': 2.*Sa })


def goodman_signal_r():
    Sm = np.array([-4., -2.,   -1., 0., 0.4, 2./3., 7./6.])
    Sa = np.array([ 2.,  2., 3./2., 1., 0.8, 2./3., 7./12.])
    warnings.simplefilter('ignore', RuntimeWarning)
    R = (Sm-Sa)/(Sm+Sa)
    warnings.simplefilter('default', RuntimeWarning)
    return pd.DataFrame({'range': 2.*Sa, 'R': R})


def five_segment_signal_sm():
    Sm = np.array([-12./5., -2., -1., 0., 2./5., 2./3., 7./6., 1.+23./75., 2.+1./150., 3.+11./25., 3.+142./225.])
    Sa = np.array([ 6./5., 2., 3./2., 1., 4./5., 2./3., 7./12., 14./25., 301./600., 86./225., 43./225.])
    return pd.DataFrame({'mean': Sm, 'range': 2.*Sa })


def five_segment_signal_r():
    Sm = np.array([-12./5., -2., -1., 0., 2./5., 2./3., 7./6., 1.+23./75., 2.+1./150., 3.+11./25., 3.+142./225.])
    Sa = np.array([ 6./5., 2., 3./2., 1., 4./5., 2./3., 7./12., 14./25., 301./600., 86./225., 43./225.])
    warnings.simplefilter('ignore', RuntimeWarning)
    R = (Sm-Sa)/(Sm+Sa)
    warnings.simplefilter('default', RuntimeWarning)
    return pd.DataFrame({'range': 2.*Sa, 'R': R })


def test_fkm_goodman_plain_sm():
    cyclic_signal = goodman_signal_sm()
    Sa = cyclic_signal['range'].to_numpy()/2.
    Sm = cyclic_signal['mean'].to_numpy()
    M = 0.5

#    R_goal = 1.
#    testing.assert_raises(ValueError, MST.fkm_goodman, Sa, Sm, M, M/3, R_goal)

    R_goal = -1.
    res = MST.fkm_goodman(Sa, Sm, M, M/3, R_goal)
    np.testing.assert_array_almost_equal(res, np.ones_like(res))

    Sm = np.array([5])
    Sa = np.array([0])
    res = MST.fkm_goodman(Sa, Sm, M, M/3, R_goal)
    assert np.equal(res, 0.)


def test_fkm_goodman_single_M_sm():
    cyclic_signal = goodman_signal_sm()
    M = 0.5

    R_goal = -1.

    res = cyclic_signal.meanstress_transform.fkm_goodman(pd.Series({'M': M, 'M2': M/3 }), R_goal).amplitude
    np.testing.assert_array_almost_equal(res, np.ones_like(res))


def test_fkm_goodman_multiple_M_sm():
    cyclic_signal = goodman_signal_sm()
    cyclic_signal.index.name = 'element_id'
    M = 0.5

    R_goal = -1.
    haigh = pd.DataFrame({'M': [M]*7, 'M2': [M/3]*7})
    res = cyclic_signal.meanstress_transform.fkm_goodman(haigh, R_goal).amplitude
    np.testing.assert_array_almost_equal(res, np.ones_like(res))


def test_five_segment_plain_sm():
    cyclic_signal = five_segment_signal_sm()
    Sa = cyclic_signal['range'].to_numpy()/2.
    Sm = cyclic_signal['mean'].to_numpy()

    M0 = 0.5
    M1 = M0/3.
    M2 = M0/6.
    M3 = 1.
    M4 = -2.

    R12 = 2./5.
    R23 = 4./5.

#    R_goal = 1.
#    testing.assert_raises(ValueError, MST.five_segment_correction, Sa, Sm, M0, M1, M2, M3, M4, R12, R23, R_goal)

    res = MST.five_segment_correction(Sa, Sm, M0=M0, M1=M1, M2=M2, M3=M3, M4=M4, R12=R12, R23=R23, R_goal=-1)
    np.testing.assert_allclose(res, np.ones_like(res))

    R_goal = -1.
    res = MST.five_segment_correction(Sa, Sm, M0, M1, M2, M3, M4, R12, R23, R_goal)
    np.testing.assert_array_almost_equal(res, np.ones_like(res))

    Sm = np.array([5])
    Sa = np.array([0])
    res = MST.five_segment_correction(Sa, Sm, M0, M1, M2, M3, M4, R12, R23, R_goal)
    assert np.equal(res, 0.)

    Sm = np.array([5, 5])
    Sa = np.array([0, 0])
    R_goal = 0.1
    res = MST.five_segment_correction(Sa, Sm, M0, M1, M2, M3, M4, R12, R23, R_goal)
    assert np.array_equal(res, np.array([0., 0.]))


def test_five_segment_single_M_sm():
    cyclic_signal = five_segment_signal_sm()
    M0 = 0.5
    M1 = M0/3.
    M2 = M0/6.
    M3 = 1.
    M4 = -2.

    R12 = 2./5.
    R23 = 4./5.

    R_goal = -1.

    res = cyclic_signal.meanstress_transform.five_segment(pd.Series({
        'M0': M0, 'M1': M1, 'M2': M2, 'M3': M3, 'M4': M4,
        'R12': R12, 'R23': R23
    }), R_goal).amplitude
    np.testing.assert_array_almost_equal(res, np.ones_like(res))


@pytest.mark.parametrize("Sm, Sa", [
    (np.array([row['mean']]), np.array([row['range']/2.])) for _, row in five_segment_signal_sm().iterrows()
])
def test_five_segment_single_M_backwards(Sm, Sa):
    cyclic_signal = pd.DataFrame({'range': [2.0], 'mean': [0.0]})
    M0 = 0.5
    M1 = M0/3.
    M2 = M0/6.
    M3 = 1.
    M4 = -2.

    R12 = 2./5.
    R23 = 4./5.

    R_goal = -np.inf if Sm+Sa == 0 else ((Sm-Sa)/(Sm+Sa))[0]

    res = cyclic_signal.meanstress_transform.five_segment(pd.Series({
        'M0': M0, 'M1': M1, 'M2': M2, 'M3': M3, 'M4': M4,
        'R12': R12, 'R23': R23
    }), R_goal)

    np.testing.assert_array_almost_equal(res.amplitude, Sa)


def test_five_segment_multiple_M_sm():
    cyclic_signal = five_segment_signal_sm()
    M0 = 0.5
    M1 = M0/3.
    M2 = M0/6.
    M3 = 1.
    M4 = -2.

    R12 = 2./5.
    R23 = 4./5.

    R_goal = -1.
    index = pd.MultiIndex.from_tuples([
        (1, 1), (1, 2), (1, 3),
        (2, 1), (2, 2), (2, 3)
    ], names=['element_id', 'node_id'])
    res = cyclic_signal.meanstress_transform.five_segment(pd.DataFrame({
        'M0': [M0]*6, 'M1': [M1]*6, 'M2': [M2]*6, 'M3': [M3]*6, 'M4': [M4]*6,
        'R12': [R12]*6, 'R23': [R23]*6
    }, index=index), R_goal)
    np.testing.assert_array_almost_equal(res.amplitude, np.ones_like(res))


#GH-105
@pytest.mark.parametrize("R_goal", [-1., 0., -1./3., 1./3.])
def test_fkm_goodman_hist_R_goal(R_goal):
    rg = pd.IntervalIndex.from_breaks(np.linspace(0., 2., 25), closed='left')
    mn = pd.IntervalIndex.from_breaks(np.linspace(-1./12., 23./12., 25), closed='left')

    mat = pd.Series(np.zeros(24*24), name='cycles',
                    index=pd.MultiIndex.from_product([rg, mn], names=['range', 'mean']))
    mat.loc[(7./6., 7./6.)] = 1.
    mat.loc[(4./3., 2./3.)] = 3.
    mat.loc[(2. - 1e-9, 0.)] = 5.

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    result = mat.meanstress_transform.fkm_goodman(haigh, R_goal).R

    expected = pd.Series(R_goal, index=result.index, name=result.name)

    pd.testing.assert_series_equal(result, expected)


@pytest.mark.parametrize("R_goal, expected", [  # all calculated by pencil on paper
    (-1., 2.0),
    (0., 4./3.),
    (-1./3., 8./5.),
    (1./3., 14./12.)
])
def test_fkm_goodman_hist_range_mean(R_goal, expected):
    rg = pd.IntervalIndex.from_breaks(np.linspace(0., 2., 25), closed='left')
    mn = pd.IntervalIndex.from_breaks(np.linspace(-1./12., 23./12., 25), closed='left')

    mat = pd.Series(np.zeros(24*24), name='cycles',
                    index=pd.MultiIndex.from_product([rg, mn], names=['range', 'mean']))
    mat.loc[(7./6., 7./6.)] = 1.
    mat.loc[(4./3., 2./3.)] = 3.
    mat.loc[(2. - 1e-9, 0.)] = 5.

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = mat.meanstress_transform.fkm_goodman(haigh, R_goal).to_pandas()
    test_interval = pd.Interval(expected-1./96., expected+1./96.)

    mask = res.index.get_level_values('range').overlaps(test_interval)
    assert res.loc[mask].sum() == 9
    assert res.loc[~mask].sum() == 0


@pytest.mark.parametrize("R_goal, expected", [  # all calculated by pencil on paper
    (-1., 2.0),
    (0., 4./3.),
    (-1./3., 8./5.),
    (1./3., 14./12.)
])
def test_fkm_goodman_hist_from_to(R_goal, expected):
    fr = pd.IntervalIndex.from_breaks(np.linspace(-25./24., 1., 49), closed='left')
    to = pd.IntervalIndex.from_breaks(np.linspace(-1./24., 2., 49), closed='left')

    mat = pd.Series(np.zeros(48*48), name='cycles',
                    index=pd.MultiIndex.from_product([fr, to], names=['from', 'to']))
    mat.loc[(14./24., 21./12.)] = 1
    mat.loc[(0., 4./3.)] = 3
    mat.loc[(-1., 1.)] = 5

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = mat.meanstress_transform.fkm_goodman(haigh, R_goal).to_pandas()

    test_interval = pd.Interval(expected-1./96., expected+1./96.)

    mask = res.index.get_level_values('range').overlaps(test_interval)
    assert res.loc[mask].sum() == 9
    assert res.loc[~mask].sum() == 0


def test_meanstress_transform_additional_index():
    fr = pd.IntervalIndex.from_breaks(np.linspace(-25./24., 1., 49), closed='left')
    to = pd.IntervalIndex.from_breaks(np.linspace(-1./24., 2., 49), closed='left')
    node = pd.Index([1, 2, 3], name='node_id')

    mat = pd.Series(
        0.0,
        index=pd.MultiIndex.from_product([fr, to, node], names=['from', 'to', 'node_id']),
        name='cycles'
    )

    mat.loc[(14./24., 21./12., 1)] = 1
    mat.loc[(0., 4./3., 1)] = 3
    mat.loc[(-1., 1., 1)] = 5

    mat.loc[(14./24., 21./12., 2)] = 2
    mat.loc[(0., 4./3., 2)] = 6
    mat.loc[(-1., 1., 2)] = 10

    mat.loc[(14./24., 21./12., 3)] = 4
    mat.loc[(0., 4./3., 3)] = 12
    mat.loc[(-1., 1., 3)] = 20

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = mat.meanstress_transform.fkm_goodman(haigh, -1.0).to_pandas()

    assert set(res.index.names) == {'range', 'mean', 'node_id'}

    null_itv = pd.Interval(0.0, 0.0)

    assert res.loc[(2.0, null_itv, 1)] == 9
    assert res.loc[(2.0, null_itv, 2)] == 18
    assert res.loc[(2.0, null_itv, 3)] == 36

    assert res.sum() == 9 + 18 + 36
    assert res.min() == 0


def test_meanstress_transform_two_additional_indices():
    fr = pd.IntervalIndex.from_breaks(np.linspace(-25./24., 1., 49), closed='left', name='from')
    to = pd.IntervalIndex.from_breaks(np.linspace(-1./24., 2., 49), closed='left', name='to')
    node = pd.Index([1, 2], name='node_id')
    element = pd.Index([1, 2], name='element_id')

    mat = pd.Series(
        0.0,
        index=pd.MultiIndex.from_product([fr, to, node, element]),
        name='cycles'
    )

    mat.loc[(14./24., 21./12., 1, 1)] = 1
    mat.loc[(0., 4./3., 1, 1)] = 3
    mat.loc[(-1., 1., 1, 1)] = 5

    mat.loc[(14./24., 21./12., 2, 1)] = 2
    mat.loc[(0., 4./3., 2, 1)] = 6
    mat.loc[(-1., 1., 2, 1)] = 10

    mat.loc[(14./24., 21./12., 1, 2)] = 4
    mat.loc[(0., 4./3., 1, 2)] = 12
    mat.loc[(-1., 1., 1, 2)] = 20

    mat.loc[(14./24., 21./12., 2, 2)] = 8
    mat.loc[(0., 4./3., 2, 2)] = 24
    mat.loc[(-1., 1., 2, 2)] = 40

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = mat.meanstress_transform.fkm_goodman(haigh, -1.0).to_pandas()

    assert set(res.index.names) == {'range', 'mean', 'node_id', 'element_id'}

    null_itv = pd.Interval(0.0, 0.0)

    assert res.loc[(2.0, null_itv, 1, 1)] == 9
    assert res.loc[(2.0, null_itv, 2, 1)] == 18

    assert res.loc[(2.0, null_itv, 1, 2)] == 36
    assert res.loc[(2.0, null_itv, 2, 2)] == 72

    assert res.sum() == 9 + 18 + 36 + 72
    assert res.min() == 0


@pytest.mark.parametrize("R_goal, expected", [  # all calculated by pencil on paper
    (-1., 2.0),
    (0., 4./3.),
    (-1./3., 8./5.),
    (1./3., 14./12.)
])
def test_fkm_goodman_hist_range_mean_nonzero(R_goal, expected):
    rg = pd.IntervalIndex.from_breaks(np.linspace(0, 2, 25), closed='left')
    mn = pd.IntervalIndex.from_breaks(np.linspace(-1./12., 23./12., 25), closed='left')

    mat = pd.Series(np.zeros(24*24), name='cycles',
                    index=pd.MultiIndex.from_product([rg, mn], names=['range', 'mean']))
    mat.loc[(7./6., 7./6.)] = 1.
    mat.loc[(4./3., 2./3.)] = 3.
    mat.loc[(2. - 1e-9, 0.)] = 5.

    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = mat[mat.values > 0].meanstress_transform.fkm_goodman(haigh, R_goal).to_pandas()

    test_interval = pd.Interval(expected-1./96., expected+1./96.)

    mask = res.index.get_level_values('range').overlaps(test_interval)
    assert res.loc[mask].sum() == 9
    assert res.loc[~mask].sum() == 0

    binsize = res.index.get_level_values('range').length.min()
    np.testing.assert_approx_equal(binsize, 2./24., significant=1)


def test_null_histogram():
    rg = pd.IntervalIndex.from_breaks(np.linspace(0, 2, 25), closed='left')
    mn = pd.IntervalIndex.from_breaks(np.linspace(0, 2, 25), closed='left')

    mat = pd.Series(np.zeros(24*24, dtype=np.int32), name='cycles',
                    index=pd.MultiIndex.from_product([rg, mn], names=['from', 'to']))
    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = mat.meanstress_transform.fkm_goodman(haigh, -1).to_pandas()

    assert not res.any()


def test_full_histogram():
    rg = pd.IntervalIndex.from_breaks(np.linspace(0, 2, 25), closed='left')
    mn = pd.IntervalIndex.from_breaks(np.linspace(0, 2, 25), closed='left')

    series = pd.Series(np.linspace(1, 576, 576, dtype=np.int32), name='cycles',
                       index=pd.MultiIndex.from_product([rg, mn], names=['from', 'to']))
    haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
    res = series.meanstress_transform.fkm_goodman(haigh, -1).to_pandas()

    assert res.sum() == series.sum()


# import itertools

# @pytest.mark.parametrize('bincount_from, bincount_to', itertools.product(*[range(1, 25), range(1, 25)]))
# def test_full_histogram_varying_bins(bincount_from, bincount_to):
#     rg = pd.IntervalIndex.from_breaks(np.linspace(0, 2, bincount_from+1), closed='left')
#     mn = pd.IntervalIndex.from_breaks(np.linspace(0, 2, bincount_to+1), closed='left')

#     bincount_prod = bincount_from * bincount_to
#     series = pd.Series(np.linspace(1, bincount_prod, bincount_prod, dtype=np.int32), name='cycles',
#                        index=pd.MultiIndex.from_product([rg, mn], names=['range', 'mean']))
#     haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
#     res = series.meanstress_transform.fkm_goodman(haigh, -1).to_pandas()

#     assert res.sum() == series.sum()


# @pytest.mark.parametrize('bincount_from, bincount_to', itertools.product(*[range(1, 25), range(1, 25)]))
# def test_full_histogram_range_mean_varying_bins(bincount_from, bincount_to):
#     fr = pd.IntervalIndex.from_breaks(np.linspace(0, 2, bincount_from+1), closed='left')
#     to = pd.IntervalIndex.from_breaks(np.linspace(0, 2, bincount_to+1), closed='left')

#     bincount_prod = bincount_from * bincount_to
#     series = pd.Series(np.linspace(1, bincount_prod, bincount_prod, dtype=np.int32), name='cycles',
#                        index=pd.MultiIndex.from_product([fr, to], names=['from', 'to']))
#     haigh = pd.Series({'M': 0.5, 'M2': 0.5/3.})
#     res = series.meanstress_transform.fkm_goodman(haigh, -1).to_pandas()

#     assert res.sum() == series.sum()


@pytest.mark.parametrize("N_c, M_sigma", [  # Calculated by pencil and paper
    (3e6, 0.2),
    (1e6, 0.3784),
    (1e5, 0.0140)
])
def test_experimental_mean_stress_sensitivity(N_c, M_sigma):
    sn_curve_R0 = pd.Series({'SD': 100, 'ND': 1e6, 'k_1': 3})
    sn_curve_Rn1 = pd.Series({'SD': 120, 'ND': 2e6, 'k_1': 5})

    testing.assert_almost_equal(
        actual=MST.experimental_mean_stress_sensitivity(sn_curve_R0, sn_curve_Rn1, N_c=N_c),
        desired=M_sigma,
        decimal=4)


def test_experimental_mean_stress_sensitivity_no_Nc():
    sn_curve_R0 = pd.Series({'SD': 100, 'ND': 1e6, 'k_1': 3})
    sn_curve_Rn1 = pd.Series({'SD': 120, 'ND': 2e6, 'k_1': 5})

    testing.assert_almost_equal(
        actual=MST.experimental_mean_stress_sensitivity(sn_curve_R0, sn_curve_Rn1),
        desired=0.2,
        decimal=4)


def test_experimental_mean_stress_sensitivity_plausible():
    sn_curve_R0 = pd.Series({'SD': 100, 'ND': 1e6, 'k_1': 3})
    sn_curve_Rn1 = pd.Series({'SD': 120, 'ND': 2e6, 'k_1': 5})
    with testing.assert_raises(ValueError):
        # Should lead to -0.27
        MST.experimental_mean_stress_sensitivity(sn_curve_R0, sn_curve_Rn1, N_c=10**4)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd
import pytest
import unittest.mock as mock


from pylife.materialdata import woehler

from .data import *


def sort_fatigue_data(fd):
    return fd.sort_values(by=['load', 'cycles']).reset_index(drop=True)


def test_fatigue_data_simple_properties():
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data
    pd.testing.assert_series_equal(fd.load.sort_values(), load_sorted)
    pd.testing.assert_series_equal(fd.cycles.sort_values(), cycles_sorted)

    assert fd.num_runouts == 18
    assert fd.num_fractures == 22


@pytest.mark.parametrize("data, finite_zone_expected, infinite_zone_expected", [
    (data,
     finite_expected,
     infinite_expected),
    (data_no_mixed_horizons,
     no_mixed_horizons_finite_expected,
     no_mixed_horizons_infinite_expected),
    (data_pure_runout_horizon_and_mixed_horizons,
     pure_runout_horizon_and_mixed_horizons_finite_expected,
     pure_runout_horizon_and_mixed_horizons_infinite_expected),
    (data_no_runouts,
     no_runouts_finite_expected,
     no_runouts_infinite_expected),
    (data_only_runout_levels,
     only_runout_levels_finite_expected,
     only_runout_levels_infinite_expected)
])
def test_fatigue_data_finite_infinite_zone(data, finite_zone_expected, infinite_zone_expected):
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.finite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(finite_zone_expected))
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.infinite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(infinite_zone_expected))

@pytest.mark.parametrize("data, finite_zone_expected, infinite_zone_expected", [
    (data,
     finite_expected_conservative,
     infinite_expected_conservative),
    (data_no_mixed_horizons,
     no_mixed_horizons_finite_expected,
     no_mixed_horizons_infinite_expected),
    (data_pure_runout_horizon_and_mixed_horizons,
     pure_runout_horizon_and_mixed_horizons_finite_expected_conservative,
     pure_runout_horizon_and_mixed_horizons_infinite_expected_conservative),
    (data_no_runouts,
     no_runouts_finite_expected,
     no_runouts_infinite_expected)
])
def test_fatigue_data_finite_infinite_zone_conservative(data, finite_zone_expected, infinite_zone_expected):
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.conservative_finite_infinite_transition()
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.finite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(finite_zone_expected))
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.conservative_finite_infinite_transition()
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.infinite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(infinite_zone_expected))

@pytest.mark.parametrize("data, finite_zone_expected, infinite_zone_expected, fat_limit", [
    (data,
     finite_expected_set_finite_limit,
     infinite_expected_set_finite_limit,
     fat_limit),
    (data,
     finite_expected_set_finite_limit_max,
     infinite_expected_set_finite_limit_max,
     fat_limit_max),
    (data,
     finite_expected_set_finite_limit_min,
     infinite_expected_set_finite_limit_min,
     fat_limit_min)
])
def test_fatigue_data_finite_infinite_zone_manual_setting(data, finite_zone_expected, infinite_zone_expected, fat_limit):
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.set_finite_infinite_transition(finite_infinite_transition=fat_limit)
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.finite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(finite_zone_expected))
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.infinite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(infinite_zone_expected))

def test_woehler_fracture_determination_given():
    df = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4]
    })

    expected = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4],
        'fracture': [True, False, True]
    })

    expected_runouts = pd.DataFrame({
        'load': [2],
        'cycles': [1e7],
        'fracture': [False]
    }, index=[1])

    expected_fractures = pd.DataFrame({
        'load': [1, 3],
        'cycles': [1e6, 1e4],
        'fracture': [True, True]
    }, index=[0, 2])

    test = woehler.determine_fractures(df, 1e7).sort_index()
    pd.testing.assert_frame_equal(test, expected)

    fd = test.fatigue_data
    pd.testing.assert_frame_equal(fd.fractures, expected_fractures)
    pd.testing.assert_frame_equal(fd.runouts, expected_runouts)


def test_woehler_fracture_determination_infered():
    df = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4]
    })

    expected = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4],
        'fracture': [True, False, True]
    })

    test = woehler.determine_fractures(df).sort_index()
    pd.testing.assert_frame_equal(test, expected)


@pytest.mark.parametrize("data, finite_infinite_transition_expected", [
    (data, 362.5),
    (data_no_mixed_horizons, 362.5),
    (data_pure_runout_horizon_and_mixed_horizons, 362.5),
    (data_no_runouts, 0.0),
    (data_only_runout_levels, 362.5)
])
def test_woehler_endur_zones(data, finite_infinite_transition_expected):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    assert fd.finite_infinite_transition == finite_infinite_transition_expected


@pytest.mark.parametrize("data, finite_infinite_transition_expected", [
    (data, 325.0),
    (data_no_mixed_horizons, 350.0),
    (data_pure_runout_horizon_and_mixed_horizons, 325.0),
    (data_no_runouts, 0.0),
    (data_only_runout_levels, 325.0)
])
def test_woehler_endur_zones_conservative(data, finite_infinite_transition_expected):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    fd = fd.conservative_finite_infinite_transition()
    assert fd.finite_infinite_transition == finite_infinite_transition_expected


def test_woehler_endure_zones_no_runouts():
    df = data[data.cycles < 1e7]
    fd = woehler.determine_fractures(df, 1e7).fatigue_data
    assert fd.finite_infinite_transition == 0.0


def test_woehler_elementary():
    expected = pd.Series({
        'SD': 362.5,
        'k_1': 9.88,
        'ND': 417772,
        'TN': 6.78,
        'TS': 1.21,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_initialize_with_determined_fractures():
    expected = pd.Series({
        'SD': 362.5,
        'k_1': 9.88,
        'ND': 417772,
        'TN': 6.78,
        'TS': 1.21,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7)
    wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_initialize_with_pandas_dataframe():
    expected = pd.Series({
        'SD': 362.5,
        'k_1': 9.88,
        'ND': 417772,
        'TN': 6.78,
        'TS': 1.21,
        'failure_probability': 0.5
    }).sort_index()

    wc = woehler.Elementary(data).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_no_runouts():
    expected = pd.Series({
        'SD': 0.0,
        'k_1': 7.0,
        'TN': 5.3,
        'TS': 1.27,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    wc = woehler.Elementary(fd).analyze().sort_index().drop('ND')
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_only_one_load_level():
    data = pd.DataFrame(np.array([[350.0, 1e7], [350.0, 1e6]]), columns=['load', 'cycles'])
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.raises(ValueError, match=r"Need at least two different load levels in the finite zone to do a Wöhler slope analysis."):
        woehler.Elementary(fd).analyze().sort_index()


def test_woehler_elementary_only_one_load_level_in_finite_region():
    expected = pd.Series({
        'k_1': np.nan,
        'ND': np.nan,
        'SD': np.nan,
        'TN': np.nan,
        'TS': np.nan,
        'failure_probability': 0.5
    }).sort_index()

    data = pd.DataFrame(np.array([[350.0, 1e7], [350.0, 1e6], [360.0, 1e6]]), columns=['load', 'cycles'])
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.warns(UserWarning, match=r"Need at least two different load levels in the finite zone to do a Wöhler slope analysis."):
        wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected)


def test_woehler_elementary_no_load_level_in_finite_region():
    expected = pd.Series({
        'k_1': np.inf,
        'ND': np.nan,
        'SD': np.nan,
        'TN': 1.0,
        'TS': np.nan,
        'failure_probability': 0.5
    }).sort_index()

    data = pd.DataFrame(np.array([[350.0, 1e7], [350.0, 1e6], [360.0, 1e7]]), columns=['load', 'cycles'])
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.warns(UserWarning, match=r"Need at least two different load levels in the finite zone to do a Wöhler slope analysis."):
        wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected)


def test_woehler_elementary_set_finite_infinite_transition_low():
    expected = pd.Series({
        'SD': 350.0,
        'k_1': 9.88,
        'ND': 590904,
        'TN': 6.78,
        'TS': 1.21,
        'failure_probability': 0.5
    }).sort_index()
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.set_finite_infinite_transition(finite_infinite_transition=350.)
    wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_set_finite_infinite_transition_high():
    expected = pd.Series({
        'SD': 376.,
        'k_1': 7.07,
        'ND': 191572,
        'TN': 4.18,
        'TS': 1.22,
        'failure_probability': 0.5
    }).sort_index()
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.set_finite_infinite_transition(finite_infinite_transition=376.)
    wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit():
    expected = pd.Series({
        'SD': 339.3,
        'TS': 1.2,
        'k_1': 9.88,
        'ND': 802898.,
        'TN': 6.78,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Probit(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit_set_finite_infinite_transition():
    expected = pd.Series({
        'SD': 340.4,
        'TS': 1.21,
        'k_1': 7.1,
        'ND': 386721.,
        'TN': 4.18,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.set_finite_infinite_transition(finite_infinite_transition=376.)
    wc = woehler.Probit(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit_one_runout_load_level():
    fd = woehler.determine_fractures(data_one_runout_load_level, 1e7).fatigue_data
    expected = woehler.Elementary(fd).analyze()
    with pytest.warns(UserWarning, match=r"Probit needs at least two – preferably mixed – load levels in the infinite zone. Falling back to Elementary."):
        wc = woehler.Probit(fd).analyze()
    pd.testing.assert_series_equal(wc, expected)


@pytest.mark.filterwarnings("error:invalid")
def test_woehler_probit_data01():
    expected = pd.Series({
        'SD': 490,
        'TS': 1.1,
        'k_1': 5.2,
        'ND': 298638.,
        'TN': 1.93,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_01, 1e7).fatigue_data
    pb = woehler.Probit(fd)
    wc = pb.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit_no_runouts():
    expected = pd.Series({
        'SD': 0.,
        'TS': 1.27,
        'k_1': 6.94,
        'ND': 4.4e30,
        'TN': 5.26,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    pb = woehler.Probit(fd)
    with pytest.warns(UserWarning):
        wc = pb.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit_no_finite_zone_data():
    expected = pd.Series({
        'SD': 340.4,
        'TS': 1.21,
        'k_1': np.inf,
        'ND': np.nan,
        'TN': 1.0,
        'failure_probability': 0.5
    }).sort_index()
    data_no_finite_zone = data[data['load']<376.].copy(deep=True)

    fd = woehler.determine_fractures(data_no_finite_zone, 1e7).sort_index().fatigue_data.set_finite_infinite_transition(finite_infinite_transition=376.)
    with pytest.warns(UserWarning, match=r"Need at least two different load levels in the finite zone to do a Wöhler slope analysis."):
        wc = woehler.Probit(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_max_likelihood_inf_limit():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.20,
        'k_1': 9.88,
        'ND': 897649.,
        'TN': 6.78,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.MaxLikeInf(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_max_likelihood_inf_limit_set_finite_infinite_transition():
    expected = pd.Series({
        'SD': 333.6,
        'TS': 1.22,
        'k_1': 7.1,
        'ND': 446054.,
        'TN': 4.18,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data.set_finite_infinite_transition(finite_infinite_transition=376.)
    wc = woehler.MaxLikeInf(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_max_likelihood_inf_limit_no_finite_zone_data():
    expected = pd.Series({
        'SD': 333.6,
        'TS': 1.22,
        'k_1': np.inf,
        'ND': np.nan,
        'TN': 1.0,
        'failure_probability': 0.5
    }).sort_index()
    data_no_finite_zone = data[data['load']<376.].copy(deep=True)

    fd = woehler.determine_fractures(data_no_finite_zone, 1e7).sort_index().fatigue_data.set_finite_infinite_transition(finite_infinite_transition=376.)
    with pytest.warns(UserWarning, match=r"Need at least two different load levels in the finite zone to do a Wöhler slope analysis."):
        wc = woehler.MaxLikeInf(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_bic_without_analysis():
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    we = woehler.MaxLikeFull(fd)
    with pytest.raises(ValueError, match="^.*BIC.*"):
        we.bayesian_information_criterion()


def test_woehler_max_likelihood_inf_limit_no_runouts():
    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    with pytest.raises(ValueError, match=r"MaxLikeHood: need at least two mixed load levels."):
        woehler.MaxLikeInf(fd).analyze().sort_index()


def test_woehler_max_likelihood_inf_limit_only_two_fractures():
    data = pd.DataFrame(np.array([[350.0, 1e7], [350.0, 5e5],[340.0, 1e7], [340.0, 1e5], [390.0, 1e4], [380.0, 1e5]]), columns=['load', 'cycles'])
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.raises(ValueError, match=r"MaxLikeHood: need at least three fractures on two load levels."):
        woehler.MaxLikeInf(fd).analyze().sort_index()


def test_woehler_max_likelihood_full_without_fixed_params():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.19,
        'k_1': 6.94,
        'ND': 463000.,
        'TN': 4.7,
        'failure_probability': 0.5
    }).sort_index()

    bic = 45.35256860035525

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    we = woehler.MaxLikeFull(fd)
    wc = we.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)
    np.testing.assert_almost_equal(we.bayesian_information_criterion(), bic, decimal=2)


def test_woehler_max_likelihood_full_without_fixed_params_no_runouts():
    expected = pd.Series({
        'SD': 0,
        'TS': 1.,
        'k_1': 6.94,
        'ND': 4.4e30,
        'TN': 5.7,
        'failure_probability': 0.5
    }).sort_index()

    bic = np.inf

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    we = woehler.MaxLikeFull(fd)
    with pytest.warns(UserWarning, match=r"^.*no runouts are present.*" ):
        wc = we.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)
    np.testing.assert_almost_equal(we.bayesian_information_criterion(), bic, decimal=2)


def test_max_likelihood_full_with_fixed_params():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.19,
        'k_1': 8.0,
        'ND': 520000.,
        'TN': 6.0,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = (
        woehler.MaxLikeFull(fd)
        .analyze(fixed_parameters={'TN': 6.0, 'k_1': 8.0})
        .sort_index()
    )
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)
    assert wc['TN'] == 6.0
    assert wc['k_1'] == 8.0


def test_max_likelihood_full_method_with_all_fixed_params():
    """
    Test of woehler curve parameters evaluation with the maximum likelihood method
    """
    fp = {'k_1': 15.7, 'TN': 1.2, 'SD': 280, 'TS': 1.2, 'ND': 10000000}
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.raises(AttributeError, match=r'You need to leave at least one parameter empty!'):
        (
            woehler.MaxLikeFull(fd)
            .analyze(fixed_parameters=fp)
        )


@pytest.mark.parametrize("data,no", [(d, i) for i, d in enumerate(all_data)])
def test_max_likelihood_parameter_sign(data, no):
    def _modify_initial_parameters_mock(fd):
        return fd

    load_cycle_limit = 1e6
    if hasattr(data, "N_threshold"):
        load_cycle_limit = data.N_threshold
    fatdat = woehler.determine_fractures(data, load_cycle_limit=load_cycle_limit)
    ml = woehler.MaxLikeFull(fatigue_data=fatdat.fatigue_data)
    wl = ml.analyze()

    print("Data set number {}".format(no))
    print("Woehler parameters: {}".format(wl))

    def assert_positive_or_nan_but_not_zero(x):
        if np.isfinite(x):
            assert x >= 0
            assert not np.isclose(x, 0.0)

    assert_positive_or_nan_but_not_zero(wl['SD'])
    assert_positive_or_nan_but_not_zero(wl['TS'])
    assert_positive_or_nan_but_not_zero(wl['k_1'])
    assert_positive_or_nan_but_not_zero(wl['ND'])
    assert_positive_or_nan_but_not_zero(wl['TN'])


@pytest.mark.parametrize("invalid_data", [data_01_one_fracture_level, data_01_two_fractures])
def test_max_likelihood_min_three_fractures_on_two_load_levels(invalid_data):
    fd = woehler.determine_fractures(invalid_data, 1e7).fatigue_data
    ml = woehler.MaxLikeFull(fatigue_data=fd)
    with pytest.raises(ValueError, match=r"^.*[N|n]eed at least.*" ):
        ml.analyze()


def test_max_likelihood_one_mixed_horizon():
    expected = pd.Series({
        'SD': 489.3,
        'TS': 1.147,
        'k_1': 7.99,
        'ND': 541e3,
        'TN': 2.51,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_01, 1e7).fatigue_data
    ml = woehler.MaxLikeFull(fatigue_data=fd)
    with pytest.warns(UserWarning, match=r"^.*less than two mixed load levels.*"):
        wc = ml.analyze().sort_index()
    bic = ml.bayesian_information_criterion()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_irrelevant_runouts_dropped():
    fd = woehler.determine_fractures(data_pure_runout_horizon_and_mixed_horizons, 1e7).fatigue_data
    num_data_before = 48
    num_tests_lowest_pure_runout_load_level = 9
    fd = fd.irrelevant_runouts_dropped()
    num_data_after = len(fd._obj)
    assert num_data_before-num_tests_lowest_pure_runout_load_level == num_data_after


def test_irrelevant_runouts_dropped_no_change():
    data_pure_runout_horizon_and_mixed_horizons
    data_extend = data_pure_runout_horizon_and_mixed_horizons.copy(deep=True)
    new_row = {'load': 2.75e+02, 'cycles': 1.00e+06}
    data_extend = pd.concat([data_extend, pd.DataFrame([new_row])])
    fd = woehler.determine_fractures(data_extend, 1e7).fatigue_data
    num_data_before = len(fd._obj)
    fd = fd.irrelevant_runouts_dropped()
    num_data_after = len(fd._obj)
    assert num_data_before == num_data_after


def test_drop_irreverent_pure_runout_levels_for_evaluation():
    expected = pd.Series({
        'SD': 339.23834,
        'TS': 1.211044,
        'k_1': 9.880429,
        'ND': 804501,
        'TN': 6.779709,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_pure_runout_horizon_and_mixed_horizons, 1e7).fatigue_data
    wc = woehler.Probit(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_drop_irreverent_pure_runout_levels_no_data_change():
    fd = woehler.determine_fractures(data_pure_runout_horizon_and_mixed_horizons, 1e7).fatigue_data
    num_data_before = len(fd._obj)
    wc = woehler.Probit(fd).analyze()
    num_data_after = len(fd._obj)
    assert num_data_before == num_data_after



# Copyright (c) 2020-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

import h5py
from h5py.h5t import string_dtype
import numpy as np

from .exceptions import *
from .vmap_dataset import VMAPDataset


class VMAPIntegrationType(VMAPDataset):
    def __init__(self, identifier, type_name, number_of_points, dimension, offset,
                 abscissas=None, weights=None, subtypes=None):
        super().__init__(identifier)
        self._type_name = type_name
        self._number_of_points = number_of_points
        self._dimension = dimension
        self._offset = offset

        self._abscissas = []
        if abscissas is not None:
            self._abscissas = abscissas

        self._weights = []
        if weights is not None:
            self._weights = weights

        self._subtypes = []
        if subtypes is not None:
            self._subtypes = subtypes

    @property
    def attributes(self):
        if self._identifier is None:
            raise (APIUseError("Need to set_identifier() before requesting the attributes."))
        return (self._identifier, self._type_name, self._number_of_points, self._dimension, self._offset,
                np.array(self._abscissas), np.array(self._weights), np.array(self._subtypes))

    @property
    def dtype(self):
        dt_type = np.dtype({"names": ["myIdentifier", "myTypeName", "myNumberOfPoints", "myDimension",
                                      "myOffset", "myAbscissas", "myWeights", "mySubTypes"],
                            "formats": ['<i4', string_dtype(), '<i4', '<i4', '<f8',
                                        h5py.special_dtype(vlen=np.dtype('float64')),
                                        h5py.special_dtype(vlen=np.dtype('float64')),
                                        h5py.special_dtype(vlen=np.dtype('float32'))]})
        return dt_type

    @property
    def dataset_name(self):
        return 'INTEGRATIONTYPES'



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility Functions
=================

A collection of functions frequently used in lifetime estimation
business.
"""

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np


def scattering_range_to_std(T):
    """Convert a scattering range (``TS`` or ``TN`` in DIN 50100:2016-12) into standard deviation.

    Parameters
    ----------
    T : float
        inverted scattering range

    Returns
    -------
    std : float
        standard deviation corresponding to TS or TN assuming a normal distribution

    Notes
    -----
    Actually ``1/(2*norm.ppf(0.9))*np.log10(T)``

    Inverse of ``std_to_scattering_range()``
    """
    return 0.39015207303618954*np.log10(T)


def std_to_scattering_range(std):
    """Convert a standard deviation into scattering range (``TS`` or ``TN`` in DIN 50100:2016-12).

    Parameters
    ----------
    std : float
        standard deviation

    Returns
    -------
    T : float
        inverted scattering range corresponding to ``std`` assuming a normal distribution

    Notes
    -----
    Actually ``10**(2*norm.ppf(0.9)*std``

    Inverse of ``scattering_range_to_std()``
    """
    return 10**(2.5631031310892007*std)


def rossow_cumfreqs(N):
    """Cumulative frequency estimator according to Rossow.

    Parameters
    ----------
    N : int
        The sample size of the statistical population

    Returns
    -------
    cumfreqs : numpy.ndarray
        The estimated cumulated frequencies of the N samples

    Notes
    -----
    The returned value is the probability that the next taken sample
    is below the value of the i-th sample of n sorted samples.

    Examples
    --------
    >>> rossow_cumfreqs(1)
    array([0.5])

    If we have one sample, the probability that the next sample will
    be below it is 0.5.

    >>> rossow_cumfreqs(3)
    array([0.2, 0.5, 0.8])

    If we have three sorted samples, the probability that the next
    sample will be
    * below the first is 0.2
    * below the second is 0.5
    * below the third is 0.8


    References
    ----------
    *Statistics of Metal Fatigue in Engineering' page 16*

    https://books.google.de/books?isbn=3752857722

    """
    i = np.arange(1, N+1)
    return (3.*i-1.)/(3.*N+1)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daniel Christopher Kreuter"
__maintainer__ = "Johannes Mueller"

import warnings

import numpy as np
import pandas as pd


def combine_histogram(hist_list, method='sum'):
    """Combine a list of histograms to one.

    Parameters
    ----------
    hist_list: list of :class:`pandas.Series`
        list of histograms with all histograms as interval indexed :class:`pandas.Series`
    method: str or aggregating function
        method used for the aggregation, e.g. 'sum', 'min', 'max', 'mean', 'std'
        or any callable function that would aggregate a :class:`pandas.Series`.
        default is 'sum'

    Returns
    -------
    histogram : :class:`pd.Series`
        The resulting histogram

    Raises
    ------
    ValueError
        if the index levels of the histograms do not match.

    Notes
    -----
    Identical bins are grouped and then aggregated using ``method``.  Note that
    neither before or after the aggregation any rebinning takes place.  You
    might consider piping your histograms through
    :func:`~pylife.utils.histogram.rebin_histogram` before or after combining them.

    The histograms need to have compatible indices. Those can either be a
    simple class:`pandas.IntervalIndex` for a one dimensional histogram or a
    :class:`pandas.MultiIndex` whose levels are all ``IntervalIndex`` for
    multidimensional histograms.  For multidimensional histograms the names of
    the index levels must match throughout the input histogram list.

    Examples
    --------
    Two one dimensional histograms:

    >>> h1 = pd.Series([5., 10.], index=pd.interval_range(start=0, end=2))
    >>> h2 = pd.Series([12., 3., 20.], index=pd.interval_range(start=1, periods=3))
    >>> h1
    (0, 1]     5.0
    (1, 2]    10.0
    dtype: float64
    >>> h2 = pd.Series([12., 3., 20.], index=pd.interval_range(start=1, periods=3))
    >>> h2
    (1, 2]    12.0
    (2, 3]     3.0
    (3, 4]    20.0
    dtype: float64
    >>> combine_histogram([h1, h2])
    (0, 1]     5.0
    (1, 2]    22.0
    (2, 3]     3.0
    (3, 4]    20.0
    dtype: float64
    >>> combine_histogram([h1, h2], method='min')
    (0, 1]     5.0
    (1, 2]    10.0
    (2, 3]     3.0
    (3, 4]    20.0
    dtype: float64
    >>> combine_histogram([h1, h2], method='max')
    (0, 1]     5.0
    (1, 2]    12.0
    (2, 3]     3.0
    (3, 4]    20.0
    dtype: float64
    >>> combine_histogram([h1, h2], method='mean')
    (0, 1]     5.0
    (1, 2]    11.0
    (2, 3]     3.0
    (3, 4]    20.0
    dtype: float64

    Limitations
    -----------
    At the moment, additional dimensions i.e. index level that are not histogram bins,
    are not supported.  This limitation might fall in the future.
    """
    def dimensions_are_consistent():
        for h in hist_list[1:]:
            if len(h.index.names) != len(hist_list[0].index.names):
                return False
            if set(h.index.names) != set(hist_list[0].index.names):
                return False

        return True

    hist_list = list(filter(lambda h: len(h) > 0, hist_list))
    if len(hist_list) == 0:
        return pd.Series(dtype=np.float64, index=pd.IntervalIndex.from_tuples([]))

    if not dimensions_are_consistent():
        raise ValueError("Histograms must have identical dimensions to be combined.")

    names = hist_list[0].index.names

    concat = pd.concat(hist_list)
    combined = concat.groupby(concat.index).agg(method)

    if isinstance(concat.index, pd.MultiIndex):
        combined.index = pd.MultiIndex.from_tuples(combined.index, names=names)

    return combined


def rebin_histogram(histogram, binning, nan_default=False):
    """Rebin a histogram to a given binning.

    Parameters
    ----------
    histogram : :class:`pandas.Series` with :class:`pandas.IntervalIndex`
        The histogram data to be rebinned

    binning : :class:`pandas.IntervalIndex` or int
        The given binning or number of bins

    nan_default : bool
        If True non occupied bins will be occupied with ``np.nan``, else 0.0
        Default False

    Returns
    -------
    rebinned : :class:`pandas.Series` with :class:`pandas.IntervalIndex`
        The rebinned histogram

    Raises
    ------
    TypeError
        if the ``histogram`` or the ``binning`` do not have an ``IntervalIndex``.
    ValueError
        if the binning is not monotonic increasing or has gaps.

    Notes
    -----
    The events collected in the bins of the original histogram are distributed
    linearly to the bins in the target bins.

    Examples
    --------
    >>> h
    (0.0, 1.0]    1.0
    (1.0, 2.0]    2.0
    (2.0, 3.0]    3.0
    (3.0, 4.0]    4.0
    dtype: float64
    >>> h = pd.Series([10.0, 20.0, 30.0, 40.0], index=pd.interval_range(0.0, 4.0, 4))
    >>> h
    (0.0, 1.0]    10.0
    (1.0, 2.0]    20.0
    (2.0, 3.0]    30.0
    (3.0, 4.0]    40.0
    dtype: float64

    Rebin to a finer binning:

    >>> target_binning = pd.interval_range(0.0, 4.0, 8)
    >>> rebin_histogram(h, target_binning)
    (0.0, 0.5]     5.0
    (0.5, 1.0]     5.0
    (1.0, 1.5]    10.0
    (1.5, 2.0]    10.0
    (2.0, 2.5]    15.0
    (2.5, 3.0]    15.0
    (3.0, 3.5]    20.0
    (3.5, 4.0]    20.0
    dtype: float64

    Rebin to a coarser binning:

    >>> target_binning = pd.interval_range(0.0, 4.0, 2)
    >>> rebin_histogram(h, target_binning)
    (0.0, 2.0]    30.0
    (2.0, 4.0]    70.0
    dtype: float64

    Define the target bin just by an int:

    >>> rebin_histogram(h, 5)
    (0.0, 0.8]     8.0
    (0.8, 1.6]    14.0
    (1.6, 2.4]    20.0
    (2.4, 3.2]    26.0
    (3.2, 4.0]    32.0
    dtype: float64

    Limitations
    -----------
    At the moment, additional dimensions i.e. index level that are not histogram bins,
    are not supported.  This limitation might fall in the future.
    """
    default_value = np.nan if nan_default else 0.0

    if not isinstance(histogram.index, pd.MultiIndex):
        return _do_rebin_histogram(histogram, binning, default_value)

    original_names = histogram.index.names
    for name in histogram.index.names:
        if not isinstance(histogram.index.get_level_values(name), pd.IntervalIndex):
            continue

        this_binning = binning.levels[binning.names.index(name)] if isinstance(binning, pd.MultiIndex) else binning

        remaining_names = list(filter(lambda m: m != name, original_names))
        histogram = (histogram
                     .groupby(remaining_names)
                     .apply(lambda h: _do_rebin_histogram(h.droplevel(remaining_names), this_binning, default_value)))

    return histogram.reorder_levels(original_names)


def _do_rebin_histogram(histogram, binning, default_value):
    def interval_overlap(reference_interval, test_interval):
        overlap = min(reference_interval.right, test_interval.right) - max(reference_interval.left, test_interval.left)
        return overlap / test_interval.length

    def aggregate_hist(interval):
        occupied = hist.loc[hist.index.overlaps(interval)].dropna()
        if len(occupied) == 0:
            return default_value

        return occupied.apply(lambda v: v.iloc[0] * interval_overlap(interval, v.name), axis=1).sum()

    def binning_of_n_bins(index, binnum):
        start = index.left.min()
        end = index.right.max()

        if np.isnan(start) or np.isnan(end):
            return pd.interval_range(0., 0., 0)

        return pd.interval_range(start, end, binnum)

    def binning_does_not_cover_histogram():
        return (
            histogram.index.right.max() > binning.right.max() or
            histogram.index.left.min() < binning.left.min()
        )

    if not isinstance(histogram.index, pd.IntervalIndex):
        raise TypeError("histogram needs to have an IntervalIndex.")

    if isinstance(binning, int):
        binning = binning_of_n_bins(histogram.index, binning)
    else:
        _fail_if_binning_invalid(binning)

    if binning_does_not_cover_histogram():
        warnings.warn("histogram is partly out of binning. This information will be lost!", RuntimeWarning)

    if len(histogram) == 0:
        rebinned = pd.Series(0.0, index=binning)
    else:
        hist = histogram.to_frame()
        rebinned = binning.to_series().apply(aggregate_hist)

    rebinned.name = histogram.name
    rebinned.index.name = histogram.index.name
    return rebinned


def _fail_if_binning_invalid(binning):
    def binning_is_overlapping_or_non_monotonic_increasing():
        return (
            len(binning) > 0
            and (
                not binning.is_non_overlapping_monotonic or binning.is_monotonic_decreasing
            )
        )

    def binning_has_gaps():
        if len(binning) == 0:
            return False
        left = binning.left[1:]
        right = binning.right[:-1]
        return pd.DataFrame({'l': left, 'r': right}).apply(lambda r: r.l != r.r, axis=1).any()

    if not isinstance(binning, pd.IntervalIndex):
        raise TypeError("binning argument must be a pandas.IntervalIndex.")

    if binning_is_overlapping_or_non_monotonic_increasing():
        raise ValueError("binning index must be monotonic increasing without overlaps.")

    if binning_has_gaps():
        raise ValueError("binning index must not have gaps.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for odbserver.

    This file was generated with PyScaffold 2.5, a tool that easily
    puts up a scaffold for your new Python project. Learn more under:
    http://pyscaffold.readthedocs.org/
"""

import sys
from setuptools import setup


def setup_package():
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    sphinx = ['sphinx'] if needs_sphinx else []
    setup(
        python_requires = "<3",
        use_scm_version = {
            "root": "../..",
            "relative_to": __file__,
            "local_scheme": "node-and-timestamp"
        },
        setup_requires=['six', 'setuptools_scm'] + sphinx
    )


if __name__ == "__main__":
    setup_package()



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
import scipy.stats as stats

from pylife.utils.functions import scattering_range_to_std

from pylife import PylifeSignal
from pylife import DataValidator


@pd.api.extensions.register_dataframe_accessor('fatigue_data')
class FatigueData(PylifeSignal):
    ''' class for fatigue data

    Mandatory keys are
        * ``load`` : float, the load level
        * ``cycles`` : float, the cycles of failure or runout
        * ``fracture``: bool, ``True`` iff the test is a fracture
     '''

    def _validate(self):
        self.fail_if_key_missing(['load', 'cycles', 'fracture'])
        self._finite_infinite_transition = None

    @property
    def num_tests(self):
        '''The number of tests'''
        return self._obj.shape[0]

    @property
    def num_fractures(self):
        '''The number of fractures'''
        return self.fractures.shape[0]

    @property
    def num_runouts(self):
        '''The number of runouts'''
        return self.runouts.shape[0]

    @property
    def fractures(self):
        '''Only the fracture tests'''
        return self._obj[self._obj.fracture]

    @property
    def runouts(self):
        '''Only the runout tests'''
        return self._obj[~self._obj.fracture]

    @property
    def load(self):
        '''The load levels'''
        return self._obj.load

    @property
    def cycles(self):
        '''the cycle numbers'''
        return self._obj.cycles

    @property
    def finite_infinite_transition(self):
        '''The start value of the load endurance limit.

        It is determined by searching for the lowest load level before the
        appearance of a runout data point, and the first load level where a
        runout appears.  Then the median of the two load levels is the start
        value.
        '''
        if self._finite_infinite_transition is None:
            self._calc_finite_infinite_transition()
        return self._finite_infinite_transition

    @property
    def finite_zone(self):
        '''All the tests with load levels above ``finite_infinite_transition``, i.e. the finite zone'''
        if self._finite_infinite_transition is None:
            self._calc_finite_infinite_transition()
        return self._finite_zone

    @property
    def infinite_zone(self):
        '''All the tests with load levels below ``finite_infinite_transition``, i.e. the infinite zone'''
        if self._finite_infinite_transition is None:
            self._calc_finite_infinite_transition()
        return self._infinite_zone

    @property
    def fractured_loads(self):
        return np.unique(self.fractures.load.values)

    @property
    def runout_loads(self):
        return np.unique(self.runouts.load.values)

    @property
    def non_fractured_loads(self):
        return np.setdiff1d(self.runout_loads, self.fractured_loads)

    @property
    def mixed_loads(self):
        return np.intersect1d(self.runout_loads, self.fractured_loads)

    @property
    def pure_runout_loads(self):
        return np.setxor1d(self.runout_loads, self.mixed_loads)

    def conservative_finite_infinite_transition(self):
        """
        Sets a lower fatigue limit that what is expected from the algorithm given by Mustafa Kassem.
        For calculating the fatigue limit, all amplitudes where runouts and fractures are present are collected.
        To this group, the maximum amplitude with only runouts present is added.
        Then, the fatigue limit is the mean of all these amplitudes.

        Returns
        -------
        self

        See also
        --------
        Kassem, Mustafa - "Open Source Software Development for Reliability and Lifetime Calculation" pp. 34
        """
        amps_to_consider = self.mixed_loads

        if len(self.non_fractured_loads ) > 0:
            amps_to_consider = np.concatenate((amps_to_consider, [self.non_fractured_loads.max()]))

        if len(amps_to_consider) > 0:
            self._finite_infinite_transition = amps_to_consider.mean()
            self._calc_finite_zone()

        return self

    def set_finite_infinite_transition(self, finite_infinite_transition):
        """
        Allows the user to set an arbitrary fatigue limit.

        Parameters
        ----------
        finite_infinite_transition : float
            The fatigue limit for separating the finite and infinite zone is set.

        Returns
        -------
        self
        """
        self._finite_infinite_transition = finite_infinite_transition
        self._calc_finite_zone_manual(finite_infinite_transition)

        return self

    def irrelevant_runouts_dropped(self):
        '''Make a copy of the instance with irrelevant pure runout levels dropped. '''
        if len(self.pure_runout_loads) <= 1:
            return self
        if self.pure_runout_loads.max() < self.fractured_loads.min():
            df = self._obj[~(self._obj.load < self.pure_runout_loads.max())]
            return FatigueData(df)
        else:
            return self

    @property
    def max_runout_load(self):
        return self.runouts.load.max()

    def _calc_finite_infinite_transition(self):
        self._calc_finite_zone()
        self._finite_infinite_transition = 0.0 if len(self.runouts) == 0 else self._half_level_above_highest_runout()

    def _half_level_above_highest_runout(self):
        if len(self._finite_zone) > 0:
            return (self._finite_zone.load.min() + self.max_runout_load) / 2.

        return self._guess_from_second_highest_runout()

    def _guess_from_second_highest_runout(self):
        max_loads = np.sort(self._obj.load.unique())[-2:]
        return max_loads[1] + (max_loads[1]-max_loads[0]) / 2.

    def _calc_finite_zone(self):
        if len(self.runouts) > 0:
            return self._calc_finite_zone_manual(self.max_runout_load)
        self._infinite_zone = self._obj[:0]
        self._finite_zone = self._obj

    def _calc_finite_zone_manual(self, limit):
        self._finite_zone = self.fractures[self.fractures.load > limit]
        self._infinite_zone = self._obj[self._obj.load <= limit]

def determine_fractures(df, load_cycle_limit=None):
    '''Adds a fracture column according to defined load cycle limit

    Parameters
    ----------
    df : DataFrame
        A ``DataFrame`` containing ``fatigue_data`` without ``fractures`` column
    load_cycle_limit : float, optional
        If given, all the tests of ``df`` with ``cycles`` equal od above
        ``load_cycle_limit`` are considered as runouts. Others as fractures.
        If not given the maximum cycle number in ``df`` is used as load cycle
        limit.

    Returns
    -------
    df : DataFrame
        A ``DataFrame`` with the column ``fracture`` added
    '''
    DataValidator().fail_if_key_missing(df, ['load', 'cycles'])
    if load_cycle_limit is None:
        load_cycle_limit = df.cycles.max()
    ret = df.copy()
    ret['fracture'] = df.cycles < load_cycle_limit
    return ret



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import pandas as pd
import numpy as np

import pylife.stress.collective


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[-1., 1.], [2, -2]]),
        pd.Series([1., 2.])
    ),
    (
        pd.DataFrame([[-2., 2.], [3, -3]], index=[23, 42]),
        pd.Series([2., 3.], index=[23, 42])
    )
])
def test_load_collective_amplitude_from_to(df, expected):
    df.columns = ['from', 'to']
    expected.name = 'amplitude'
    pd.testing.assert_series_equal(df.load_collective.amplitude, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.cycles, expected_cycles)


def test_load_collective_amplitude_from_to_with_cycles():
    df = pd.DataFrame([[-1., 1., 1e6], [2, -2, 2e6]])
    df.columns = ['from', 'to', 'cycles']

    expected_cycles = pd.Series([1e6, 2e6], name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.cycles, expected_cycles)


def test_load_collective_upper_from_to_with_cycles():
    df = pd.DataFrame([[-1., 1., 1e6], [2, -2, 2e6]])
    df.columns = ['from', 'to', 'cycles']

    expected_upper = pd.Series([1., 2.], name='upper', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.upper, expected_upper)


def test_load_collective_lower_from_to_with_cycles():
    df = pd.DataFrame([[1000., 10000., 1e2], [20000., 2000., 2e2]])
    df.columns = ['from', 'to', 'cycles']

    expected_lower = pd.Series([1000., 2000.], name='lower', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.lower, expected_lower)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[-1., 1.], [3, -1]]),
        pd.Series([0., 1.])
    ),
    (
        pd.DataFrame([[-2., 2.], [6, -2]], index=[23, 42]),
        pd.Series([0., 2.], index=[23, 42])
    )
])
def test_load_collective_mean_from_to(df, expected):
    df.columns = ['from', 'to']
    expected.name = 'meanstress'
    pd.testing.assert_series_equal(df.load_collective.meanstress, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.cycles, expected_cycles)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([
            [-1., 1.],
            [3, -1],
            [0., 0.],
            [-1.0, 0.0],
        ]),
        pd.Series([-1., -1./3., 0., -np.inf])
    ),
    (
        pd.DataFrame([[-2., 2.], [6, -2]], index=[23, 42]),
        pd.Series([-1., -1./3.], index=[23, 42])
    )
])
def test_load_collective_R_from_to(df, expected):
    df.columns = ['from', 'to']
    expected.name = 'R'
    pd.testing.assert_series_equal(df.load_collective.R, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.cycles, expected_cycles)


@pytest.mark.parametrize('df, expected_upper, expected_lower', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[-1., 1.], [3, -1]]),
        pd.Series([1., 3.]), pd.Series([-1., -1])
    ),
    (
        pd.DataFrame([[-2., 2.], [6, -2]], index=[23, 42]),
        pd.Series([2., 6.], index=[23, 42]),
        pd.Series([-2., -2.], index=[23, 42]),
    )
])
def test_load_collective_upper_lower_from_to(df, expected_upper, expected_lower):
    df.columns = ['from', 'to']
    expected_upper.name = 'upper'
    expected_lower.name = 'lower'
    pd.testing.assert_series_equal(df.load_collective.upper, expected_upper)
    pd.testing.assert_series_equal(df.load_collective.lower, expected_lower)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
    ),
    (
        pd.DataFrame([[-1., 1.], [2, -2]]),
        pd.DataFrame([[-2., 2.], [4, -4]]),
    ),
    (
        pd.DataFrame([[-2., 2.], [3, -3]], index=[23, 42]),
        pd.DataFrame([[-4., 4.], [6, -6]], index=[23, 42]),
    )
])
def test_load_collective_from_to_scale_scalar(df, expected):
    df.columns = ['from', 'to']
    expected.columns = ['from', 'to']
    pd.testing.assert_frame_equal(df.load_collective.scale(2.0).to_pandas(), expected)


def test_load_collective_from_to_scale_scalar_with_cycles():
    df = pd.DataFrame([[-1., 1., 1e6], [2, -2, 2e6]])
    df.columns = ['from', 'to', 'cycles']
    expected = pd.DataFrame([[-2., 2., 1e6], [4, -4, 2e6]])
    expected.columns = ['from', 'to', 'cycles']

    pd.testing.assert_frame_equal(df.load_collective.scale(2.0).to_pandas(), expected)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
    ),
    (
        pd.DataFrame([[-1., 1.], [2, -2]]),
        pd.DataFrame([[1., 3.], [4., 0.]]),
    ),
    (
        pd.DataFrame([[-2., 2.], [3., -3.]], index=[23, 42]),
        pd.DataFrame([[0., 4.], [5., -1.]], index=[23, 42]),
    )
])
def test_load_collective_from_to_shift_scalar(df, expected):
    df.columns = ['from', 'to']
    expected.columns = ['from', 'to']
    pd.testing.assert_frame_equal(df.load_collective.shift(2.0).to_pandas(), expected)


def test_load_collective_from_to_shift_scalar_with_cycles():
    df = pd.DataFrame([[-1., 1., 1e2], [2, -2, 2e2]])
    df.columns = ['from', 'to', 'cycles']
    expected = pd.DataFrame([[1., 3., 1e2], [4., 0., 2e2]])
    expected.columns = ['from', 'to', 'cycles']
    pd.testing.assert_frame_equal(df.load_collective.shift(2.0).to_pandas(), expected)


def test_load_collective_from_to_scale_series():
    df = pd.DataFrame([[-2., 2.], [3, -3]],
                      columns=['from', 'to'],
                      index=pd.Index([23, 42], name='foo_index'))

    scale_operand = pd.Series([2, 3, 4], index=pd.Index([6, 7, 8], name='scale_index'))

    expected_index = pd.MultiIndex.from_tuples([
        (23, 6), (23, 7), (23, 8),
        (42, 6), (42, 7), (42, 8),
    ], names=['foo_index', 'scale_index'])
    expected = pd.DataFrame([
        [-4., 4.],
        [-6., 6.],
        [-8., 8.],
        [6., -6.],
        [9., -9.],
        [12., -12.]
    ], columns=['from', 'to'], index=expected_index)

    pd.testing.assert_frame_equal(df.load_collective.scale(scale_operand).to_pandas(), expected)


def test_load_collective_from_to_shift_series():
    df = pd.DataFrame([[-2., 2.], [3, -3]],
                      columns=['from', 'to'],
                      index=pd.Index([23, 42], name='foo_index'))

    shift_operand = pd.Series([2, 3, 4], index=pd.Index([6, 7, 8], name='shift_index'))

    expected_index = pd.MultiIndex.from_tuples([
        (23, 6), (23, 7), (23, 8),
        (42, 6), (42, 7), (42, 8),
    ], names=['foo_index', 'shift_index'])
    expected = pd.DataFrame([
        [0., 4.],
        [1., 5.],
        [2., 6.],
        [5., -1.],
        [6., 0.],
        [7., 1.]
    ], columns=['from', 'to'], index=expected_index)

    pd.testing.assert_frame_equal(df.load_collective.shift(shift_operand).to_pandas(), expected)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([2., 1.])
    ),
    (
        pd.DataFrame([[6., 2.], [4, -3]], index=[23, 42]),
        pd.Series([3., 2.], index=[23, 42])
    )
])
def test_load_collective_amplitude_range_mean(df, expected):
    df.columns = ['range', 'mean']
    expected.name = 'amplitude'
    pd.testing.assert_series_equal(df.load_collective.amplitude, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.cycles, expected_cycles)


def test_load_collective_amplitude_range_mean_with_cycles():
    df = pd.DataFrame([[2., 1., 1e6], [0.1, 0.2, 2e6]])
    df.columns = ['range', 'mean', 'cycles']

    expected_cycles = pd.Series([1e6, 2e6], name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.load_collective.cycles, expected_cycles)


@pytest.mark.parametrize('df, expected_upper, expected_lower', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([3., 1.]), pd.Series([-1., -1])
    ),
    (
        pd.DataFrame([[6., 2.], [4, -3]], index=[23, 42]),
        pd.Series([5., -1.], index=[23, 42]),
        pd.Series([-1., -5.], index=[23, 42]),
    )
])
def test_load_collective_upper_lower_range_mean(df, expected_upper, expected_lower):
    df.columns = ['range', 'mean']
    expected_upper.name = 'upper'
    expected_lower.name = 'lower'
    pd.testing.assert_series_equal(df.load_collective.upper, expected_upper)
    pd.testing.assert_series_equal(df.load_collective.lower, expected_lower)


def test_load_collective_upper_lower_range_mean_single_value():
    df = pd.DataFrame({'range': [2.0], 'mean': 0.0})
    expected = pd.DataFrame({'from': [-1.0], 'to': [1.0]})
    pd.testing.assert_frame_equal(df.load_collective.to_pandas(), expected)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([1., 0.])
    ),
    (
        pd.DataFrame([[6., 2.], [4, -3]], index=[23, 42]),
        pd.Series([2., -3.], index=[23, 42])
    )
])
def test_load_collective_mean_range_mean(df, expected):
    df.columns = ['range', 'mean']
    expected.name = 'meanstress'
    pd.testing.assert_series_equal(df.load_collective.meanstress, expected)


@pytest.mark.parametrize('df, expected_amplitude, expected_mean', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([4., 2.]), pd.Series([2., 0.])
    ),
    (
        pd.DataFrame([[6., 2.], [4., -3.]], index=[23, 42]),
        pd.Series([6., 4.], index=[23, 42]),
        pd.Series([4., -6.], index=[23, 42]),
    )
])
def test_load_collective_mean_range_scale_scalar(df, expected_amplitude, expected_mean):
    df.columns = ['range', 'mean']
    expected_amplitude.name = 'amplitude'
    expected_mean.name = 'meanstress'
    scaled = df.load_collective.scale(2.0)
    pd.testing.assert_series_equal(scaled.amplitude, expected_amplitude)
    pd.testing.assert_series_equal(scaled.meanstress, expected_mean)


@pytest.mark.parametrize('df, expected_amplitude, expected_mean', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([2., 1.]), pd.Series([3., 2.])
    ),
    (
        pd.DataFrame([[6., 2.], [4., -3.]], index=[23, 42]),
        pd.Series([3., 2.], index=[23, 42]),
        pd.Series([4., -1.], index=[23, 42]),
    )
])
def test_load_collective_mean_range_shift_scalar(df, expected_amplitude, expected_mean):
    df.columns = ['range', 'mean']
    expected_amplitude.name = 'amplitude'
    expected_mean.name = 'meanstress'
    scaled = df.load_collective.shift(2.0)
    pd.testing.assert_series_equal(scaled.amplitude, expected_amplitude)
    pd.testing.assert_series_equal(scaled.meanstress, expected_mean)


@pytest.mark.parametrize('bins, expected_index_tuples, expected_data', [
    ([0, 1, 2, 3], [(0, 1), (1, 2), (2, 3)], [0, 2, 1]),
    ([1, 2, 3], [(1, 2), (2, 3)], [2, 1])
])
def test_load_collective_range_histogram_alter_bins(bins, expected_index_tuples, expected_data):
    df = pd.DataFrame({
        'range': [1., 2., 1.],
        'mean': [0, 0, 0]
    }, columns=['range', 'mean'])

    expected = pd.Series(expected_data,
                         name='cycles',
                         index=pd.IntervalIndex.from_tuples(expected_index_tuples, name='range'))

    result = df.load_collective.range_histogram(bins)

    pd.testing.assert_series_equal(result.to_pandas(), expected)


def test_load_collective_range_histogram_alter_ranges():
    df = pd.DataFrame({
        'range': [1., 2., 1., 2., 1],
        'mean': [0, 0, 0, 0, 0]
    }, columns=['range', 'mean'])

    expected_index = pd.IntervalIndex.from_tuples([(0, 1), (1, 2), (2, 3)], name='range')
    expected = pd.Series([0, 3, 2], name='cycles', index=expected_index)

    result = df.load_collective.range_histogram([0, 1, 2, 3])

    pd.testing.assert_series_equal(result.to_pandas(), expected)


def test_load_collective_range_histogram_interval_index():
    df = pd.DataFrame({
        'range': [1., 2., 1., 2., 1],
        'mean': [0, 0, 0, 0, 0]
    }, columns=['range', 'mean'])

    expected_index = pd.IntervalIndex.from_tuples([(0, 1), (1, 2), (2, 3)], name='range')
    expected = pd.Series([0, 3, 2], name='cycles', index=expected_index)

    result = df.load_collective.range_histogram(expected_index)

    pd.testing.assert_series_equal(result.to_pandas(), expected)


def test_load_collective_range_histogram_unnested_grouped():
    element_idx = pd.Index([10, 20, 30], name='element_id')
    cycle_idx = pd.Index([0, 1, 2], name='cycle_number')
    idx = pd.MultiIndex.from_product((element_idx, cycle_idx))

    df = pd.DataFrame({
        'range': [1., 2., 1., 2., 1., 2., 1., 1., 1],
        'mean': [0, 0, 0, 0, 0, 0, 0, 0, 0]
    }, columns=['range', 'mean'], index=idx)

    expected_intervals = pd.IntervalIndex.from_tuples([(0, 1), (1, 2), (2, 3)], name='range')
    expected_index = pd.MultiIndex.from_product([element_idx, expected_intervals])
    expected = pd.Series([0, 2, 1, 0, 1, 2, 0, 3, 0], name='cycles', index=expected_index)

    result = df.load_collective.range_histogram([0, 1, 2, 3], 'cycle_number')

    pd.testing.assert_series_equal(result.to_pandas(), expected)


def test_load_collective_range_histogram_nested_grouped():
    element_idx = pd.Index([10, 20], name='element_id')
    node_idx = pd.Index([100, 101], name='node_id')
    cycle_idx = pd.Index([0, 1], name='cycle_number')
    idx = pd.MultiIndex.from_product((element_idx, node_idx, cycle_idx))

    df = pd.DataFrame({
        'range': [1., 2., 1., 2., 1., 2., 1., 1.],
        'mean': [0, 0, 0, 0, 0, 0, 0, 0]
    }, columns=['range', 'mean'], index=idx)

    expected_intervals = pd.IntervalIndex.from_tuples([(0, 1), (1, 2), (2, 3)], name='range')
    expected_index = pd.MultiIndex.from_product([element_idx, node_idx, expected_intervals])
    expected = pd.Series([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 2, 0], name='cycles', index=expected_index)

    result = df.load_collective.range_histogram([0, 1, 2, 3], 'cycle_number')

    pd.testing.assert_series_equal(result.to_pandas(), expected)


def test_load_collective_strange_shift():
    upper_loads = pd.Series([1000., 2000., 1500])
    collective = pd.DataFrame({
        'to': upper_loads,
        'from': 0.0,
    })
    collective.index.name = 'load_block'

    mises = pd.Series([1.0, 2.0, 3.0], index=pd.Index([1, 2, 3], name='node_id'), name='mises')

    lower_stress = pd.Series([100., 200., 300.], index=mises.index, name='lower_stress')

    result = collective.load_collective.scale(mises).shift(lower_stress).to_pandas()

    expected_index = pd.MultiIndex.from_product([collective.index, mises.index])
    expected = pd.DataFrame({
        'to': [1100., 2200., 3300., 2100., 4200., 6300., 1600., 3200., 4800.],
        'from': [100., 200., 300.] * 3
    }, index=expected_index)

    pd.testing.assert_frame_equal(result, expected)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd
import numpy as np

from pylife import PylifeSignal


@pd.api.extensions.register_dataframe_accessor("voigt")
class StressTensorVoigt(PylifeSignal):
    '''DataFrame accessor class for Voigt noted stress tensors

    Raises
    ------
    AttributeError
        if at least one of the needed columns is missing.

    Notes
    -----
    Base class to access :class:`pandas.DataFrame` objects containing
    Voigt noted stress tensors. The stress tensor components are assumed
    to be in the columns `S11`, `S22`, `S33`, `S12`, `S13`, `S23`.

    See also
    --------
    :func:`pandas.api.extensions.register_dataframe_accessor()`

    Examples
    --------
    For an example see :class:`equistress.StressTensorEquistress`.
    '''
    def _validate(self):
        self.fail_if_key_missing(['S11', 'S22', 'S33', 'S12', 'S13', 'S23'])



import re

import numpy as np
import pandas as pd
import pytest

import pylife.vmap as vmap

from . import reference_data as RD


@pytest.fixture
def beam_2d_squ():
    return vmap.VMAPImport('tests/vmap/testfiles/beam_2d_squ_lin.vmap')


@pytest.fixture
def beam_3d_hex():
    return vmap.VMAPImport('tests/vmap/testfiles/beam_3d_hex_lin.vmap')


@pytest.fixture
def rotsym_quad():
    return vmap.VMAPImport('tests/vmap/testfiles/rotsym_quad_lin.vmap')


@pytest.fixture
def beam_2d_squ_lin_and_quad():
    return vmap.VMAPImport('tests/vmap/testfiles/beam_2d_squ_lin_and_quad.vmap')


def assert_list_equal(l1, l2):
    assert len(l1) == len(l2)
    for e in l1:
        assert e in l2


def test_get_geometries(beam_2d_squ_lin_and_quad):
    assert_list_equal(beam_2d_squ_lin_and_quad.geometries(), ['1', '2'])


def test_get_states(beam_2d_squ):
    assert_list_equal(beam_2d_squ.states(), ['STATE-1', 'STATE-2'])


def test_get_nodes_2d(beam_2d_squ):
    pd.testing.assert_frame_equal(beam_2d_squ.nodes('1'), RD.beam_2d_squ_nodes)


def test_make_mesh_no_set(beam_2d_squ):
    df = beam_2d_squ.make_mesh('1').to_frame()
    assert df.shape[1] == 0
    pd.testing.assert_index_equal(df.index, RD.beam_2d_squ_mesh_coords.index)


def test_make_mesh_unknown_geometry_one_geometry(beam_2d_squ):
    with pytest.raises(KeyError, match=re.escape("Geometry 'foo' not found. "
                                                   "Available geometries: ['1'].")):
        beam_2d_squ.make_mesh('foo')


def test_make_mesh_unknown_geometry_two_geometries(beam_2d_squ_lin_and_quad):
    with pytest.raises(KeyError, match=re.escape("Geometry 'foo' not found. "
                                                   "Available geometries: ['1', '2'].")):
        beam_2d_squ_lin_and_quad.make_mesh('foo')


def test_get_make_mesh_beam_2d_sq_node_set_load(beam_2d_squ):
    pd.testing.assert_index_equal(beam_2d_squ.make_mesh('1').filter_node_set('LOAD').to_frame().index,
                                  RD.beam_2d_squ_mesh_index_load)


def test_get_make_mesh_beam_2d_sq_node_set_fix(beam_2d_squ):
    pd.testing.assert_index_equal(beam_2d_squ.make_mesh('1').filter_node_set('FIX').to_frame().index,
                                  RD.beam_2d_squ_mesh_index_fix)


def test_get_make_mesh_rotsym_quad_element_set_ysym(rotsym_quad):
    pd.testing.assert_index_equal(rotsym_quad.make_mesh('1').filter_element_set('YSYM').to_frame().index,
                                  RD.rotsym_quad_mesh_index_ysym)


def test_make_mesh_join_coordinates_filtered_node_set(beam_2d_squ):
    pd.testing.assert_frame_equal(beam_2d_squ.make_mesh('1')
                                  .filter_node_set('ALL')
                                  .join_coordinates().to_frame(),
                                  RD.beam_2d_squ_mesh_coords)


def test_make_mesh_join_coordinates_filtered_element_set(beam_2d_squ):
    pd.testing.assert_frame_equal(beam_2d_squ.make_mesh('1')
                                  .filter_element_set('ALL')
                                  .join_coordinates().to_frame(),
                                  RD.beam_2d_squ_mesh_coords)


def test_make_mesh_join_coordinates_unfiltered(beam_2d_squ):
    pd.testing.assert_frame_equal(beam_2d_squ.make_mesh('1').join_coordinates().to_frame(),
                                  RD.beam_2d_squ_mesh_coords)


def test_make_mesh_join_coordinates_no_mesh(beam_2d_squ):
    with pytest.raises(vmap.APIUseError, match=re.escape("Need to make_mesh() before joining the coordinates.")):
        beam_2d_squ.join_coordinates()


def test_make_mesh_join_coordinates_node_set_load(rotsym_quad):
    pd.testing.assert_frame_equal(rotsym_quad.make_mesh('1').filter_element_set('YSYM').join_coordinates().to_frame(),
                                  RD.rotsym_quad_mesh_coords_ysym.loc[RD.rotsym_quad_mesh_index_ysym])


def test_make_mesh_join_coordinates_element_set_ysym(beam_2d_squ):
    pd.testing.assert_frame_equal(beam_2d_squ.make_mesh('1').filter_node_set('LOAD').join_coordinates().to_frame(),
                                  RD.beam_2d_squ_mesh_coords.loc[RD.beam_2d_squ_mesh_index_load])


def test_get_make_mesh_fail_nonexistant_node_set(beam_2d_squ):
    with pytest.raises(KeyError, match="Node set 'foo' not found in geometry '1'"):
        beam_2d_squ.make_mesh('1').filter_node_set('foo')


def test_get_make_mesh_fail_nonexistant_element_set(beam_2d_squ):
    with pytest.raises(KeyError, match="Element set 'foo' not found in geometry '1'"):
        beam_2d_squ.make_mesh('1').filter_element_set('foo')


def test_filter_node_set_no_mesh(beam_2d_squ):
    with pytest.raises(vmap.APIUseError, match=re.escape("Need to make_mesh() before filtering node or element sets.")):
        beam_2d_squ.filter_node_set('FIX')


def test_filter_element_set_no_mesh(beam_2d_squ):
    with pytest.raises(vmap.APIUseError, match=re.escape("Need to make_mesh() before filtering node or element sets.")):
        beam_2d_squ.filter_element_set('YSYM')


def test_to_frame_no_mesh(beam_2d_squ):
    with pytest.raises(vmap.APIUseError, match=re.escape("Need to make_mesh() before requesting a resulting frame.")):
        beam_2d_squ.to_frame()


def test_to_frame_twice(beam_2d_squ):
    vm = beam_2d_squ.make_mesh('1')
    vm.to_frame()
    with pytest.raises(vmap.APIUseError, match=re.escape("Need to make_mesh() before requesting a resulting frame.")):
        vm.to_frame()


def test_join_variable_no_mesh(beam_2d_squ):
    with pytest.raises(vmap.APIUseError, match=re.escape("Need to make_mesh() before joining a variable.")):
        beam_2d_squ.join_variable('STATE-1', 'STRESS_CAUCHY')


def test_available_variables(beam_2d_squ_lin_and_quad):
    np.testing.assert_array_equal(beam_2d_squ_lin_and_quad.variables('1', 'STATE-1'),
                                  ['FORCE_CONCENTRATED', 'E', 'FORCE_REACTION', 'STRESS_CAUCHY', 'DISPLACEMENT'])
    np.testing.assert_array_equal(beam_2d_squ_lin_and_quad.variables('1', 'STATE-2'),
                                  ['FORCE_CONCENTRATED', 'E', 'STRESS_CAUCHY', 'DISPLACEMENT'])
    np.testing.assert_array_equal(beam_2d_squ_lin_and_quad.variables('2', 'STATE-1'),
                                  ['E', 'FORCE_REACTION', 'STRESS_CAUCHY', 'DISPLACEMENT'])


def test_available_variables_unknown_geometry(beam_2d_squ_lin_and_quad):
    with pytest.raises(KeyError, match=re.escape("Geometry")):
        beam_2d_squ_lin_and_quad.variables('foo', 'STATE-1')


def test_available_variables_unknown_state(beam_2d_squ_lin_and_quad):
    with pytest.raises(KeyError, match=re.escape("State")):
        beam_2d_squ_lin_and_quad.variables('1', 'state-foo')


def test_available_variables_unknown_geometry_in_state(beam_2d_squ_lin_and_quad):
    with pytest.raises(KeyError, match=re.escape("Geometry '2' not available in state 'STATE-2'.")):
        beam_2d_squ_lin_and_quad.variables('2', 'STATE-2')


def test_join_node_variable_displacement(beam_2d_squ):
    var_frame = beam_2d_squ.make_mesh('1').join_variable('DISPLACEMENT', 'STATE-2').to_frame()
    groups = var_frame.groupby('node_id')
    pd.testing.assert_frame_equal(groups.mean(), RD.beam_2d_squ_node_displacement, check_index_type=False)
    pd.testing.assert_frame_equal(groups.min(), RD.beam_2d_squ_node_displacement, check_index_type=False)
    pd.testing.assert_frame_equal(groups.max(), RD.beam_2d_squ_node_displacement, check_index_type=False)


def test_join_variable_unknown_state(beam_2d_squ):
    with pytest.raises(KeyError, match=re.escape("State 'foo' not found. "
                                                 "Available states: ['STATE-1', 'STATE-2'].")):
        beam_2d_squ.make_mesh('1').join_variable('DISPLACEMENT', 'foo')


def test_join_variable_unknown_variable(beam_2d_squ):
    with pytest.raises(KeyError, match=re.escape("Variable 'foo' not found in geometry '1', 'STATE-1'.")):
        beam_2d_squ.make_mesh('1').join_variable('foo', 'STATE-1', column_names=['foo'])


def test_join_element_variable_evol_no_column_name(beam_3d_hex):
    with pytest.raises(KeyError, match="No column name for variable EVOL. Please provide with column_names parameter"):
        beam_3d_hex.make_mesh('1').join_variable('EVOL', 'STATE-2')


def test_join_element_variable_evol(beam_3d_hex):
    var_frame = beam_3d_hex.make_mesh('1').join_variable('EVOL', 'STATE-2', column_names=['Ve']).to_frame()
    groups = var_frame.groupby('element_id')
    pd.testing.assert_frame_equal(groups.mean(), RD.beam_3d_hex_element_volume)
    pd.testing.assert_frame_equal(groups.min(), RD.beam_3d_hex_element_volume)
    pd.testing.assert_frame_equal(groups.max(), RD.beam_3d_hex_element_volume)


def test_join_element_nodal_variable_stress(beam_2d_squ):
    var_frame = beam_2d_squ.make_mesh('1').join_variable('STRESS_CAUCHY', 'STATE-2').to_frame()
    pd.testing.assert_frame_equal(var_frame, RD.beam_2d_squ_element_nodal_stress)


def test_join_element_nodal_variable_stress_override_column_names(beam_2d_squ):
    var_frame = (beam_2d_squ.make_mesh('1').join_variable('STRESS_CAUCHY', 'STATE-2',
                                                          column_names=['s11', 's22', 's33', 's12', 's13', 's23'])
                 .to_frame())
    reference = RD.beam_2d_squ_element_nodal_stress.copy()
    reference.columns = ['s11', 's22', 's33', 's12', 's13', 's23']
    pd.testing.assert_frame_equal(var_frame, reference)


def test_join_element_nodal_variable_stress_column_name_list_unmatch(beam_2d_squ):
    with pytest.raises(ValueError,
                       match=re.escape("Length of column name list (3) does not match variable dimension (6).")):
        beam_2d_squ.make_mesh('1').join_variable('STRESS_CAUCHY', 'STATE-2', column_names=['s11', 's22', 's33'])


def test_join_element_nodal_variable_strain(beam_2d_squ):
    var_frame = beam_2d_squ.make_mesh('1').join_variable('E', 'STATE-2').to_frame()
    pd.testing.assert_frame_equal(var_frame, RD.beam_2d_squ_element_nodal_strain)


def test_join_element_nodal_variable_node_set(beam_2d_squ):
    pd.testing.assert_frame_equal(beam_2d_squ
                                  .make_mesh('1')
                                  .filter_node_set('FIX')
                                  .join_variable('STRESS_CAUCHY', 'STATE-2')
                                  .to_frame(),
                                  RD.beam_2d_squ_element_nodal_stress.loc[RD.beam_2d_squ_mesh_index_fix])


def test_join_element_nodal_variable_element_set(rotsym_quad):
    pd.testing.assert_frame_equal(rotsym_quad.make_mesh('1')
                                  .filter_element_set('YSYM')
                                  .join_variable('STRESS_CAUCHY', 'STATE-2')
                                  .to_frame(),
                                  RD.rotsym_quad_stress_cauchy.loc[RD.rotsym_quad_mesh_index_ysym])


def test_join_element_nodal_variable_stress_element_variable_evol(beam_3d_hex):
    var_frame = (beam_3d_hex
                 .make_mesh('1')
                 .join_variable('STRESS_CAUCHY', 'STATE-2')
                 .join_variable('EVOL', column_names=['V_e'])
                 .to_frame())
    pd.testing.assert_frame_equal(var_frame, RD.beam_3d_hex_stress_element_volume, rtol=1e-4, check_index_type=False)


def test_join_variable_unsupported_location():
    vm = vmap.VMAPImport('tests/vmap/testfiles/beam_at_integration_points.vmap').make_mesh('1')
    with pytest.raises(vmap.FeatureNotSupportedError,
                       match="Unsupported value location, sorry\nSupported: NODE, ELEMENT, ELEMENT NODAL"):
        vm.join_variable('STRESS_CAUCHY', 'STATE-2')


def test_get_element_nodal_variable_strain_predefined_state(beam_2d_squ):
    var_frame = beam_2d_squ.make_mesh('1', 'STATE-2').join_variable('E').to_frame()
    pd.testing.assert_frame_equal(var_frame, RD.beam_2d_squ_element_nodal_strain)


def test_get_variables_multiple_states(beam_2d_squ):
    var_frame = (beam_2d_squ.make_mesh('1')
                 .join_variable('STRESS_CAUCHY', 'STATE-1', column_names=['S11_0', 'S22_0', 'S33_0', 'S12_0', 'S13_0', 'S23_0'])
                 .join_variable('E', 'STATE-2')
                 .join_variable('STRESS_CAUCHY')
                 .to_frame())
    strain = RD.beam_2d_squ_element_nodal_strain
    stress = RD.beam_2d_squ_element_nodal_stress
    zero_stress = pd.DataFrame(np.zeros((strain.shape[0], 6)),
                               columns=['S11_0', 'S22_0', 'S33_0', 'S12_0', 'S13_0', 'S23_0'],
                               index=strain.index)
    pd.testing.assert_frame_equal(var_frame, zero_stress.join(strain).join(stress))


def test_join_variable_no_given_state(beam_2d_squ):
    with pytest.raises(vmap.APIUseError,
                       match=re.escape("No state name given.\n"
                                       "Must be either given in make_mesh() or in join_variable() as optional state argument.")):
        beam_2d_squ.make_mesh('1').join_variable('STRESS_CAUCHY')


def test_get_node_sets_beam(beam_2d_squ):
    assert_list_equal(beam_2d_squ.node_sets('1'), ['ALL', 'FIX', 'LOAD'])


def test_get_node_sets_rotsym(rotsym_quad):
    assert_list_equal(rotsym_quad.node_sets('1'), ['ALL', 'LOAD', 'YSYM', 'ROTSYM', 'SLOPEDSURFACE'])


def test_get_element_sets_beam(beam_2d_squ):
    assert_list_equal(beam_2d_squ.element_sets('1'), ['ALL'])


def test_get_element_sets_rotsym(rotsym_quad):
    assert_list_equal(rotsym_quad.element_sets('1'), ['ALL', 'LOAD', 'YSYM', 'ROTSYM', 'SLOPEDSURFACE', 'LOADSURFACE'])


def test_get_mesh_index_fail_nonexistant_node_set(beam_2d_squ):
    with pytest.raises(KeyError, match="Node set 'foo' not found in geometry '1'"):
        beam_2d_squ.make_mesh('1').filter_node_set('foo')


def test_get_mesh_index_fail_nonexistant_element_set(beam_2d_squ):
    with pytest.raises(KeyError, match="Element set 'foo' not found in geometry '1'"):
        beam_2d_squ.make_mesh('1').filter_element_set('foo')
